//############################################################################
//
// LaserBoy !!!
//
// by James Lehman
// Extra Stimulus Inc.
// james@akrobiz.com
//
// began: October 2003
//
//############################################################################
#include "ez_uail_gui.hpp"
#include "ez_wave.hpp"

//############################################################################
ez_uail_gui* AIL    = NULL; // the one and only!
SDL_Surface* screen = NULL; // an SDL graphics window
SDL_Event    event        ;

//############################################################################
ez_uail_gui::ez_uail_gui(u_int xres, u_int yres)
                : ez_uail_app             (          )
                , bmp                     (          )
                , background              (          )
                , background_bmp          (          )
                , bg_file_name            ("."       )
                , bg_color                (0, 0, 0   ) // black
                , move_points_per_key     (100       ) // pxls
                , view_angle              (          )
                , view_scale              (1.0       )
                , view_offset             (          )
                , rotate_radians_per_key  (one_degree) // radians
                , scale_per_key           (0.01      )
                , scale_up_per_key        (1.01      )
                , scale_dn_per_key        (0.99      )
                , show_points             (false     )
                , show_blanking           (false     )
                , show_cursors            (false     )
                , show_palettes           (true      )
                , show_stats              (true      )
                , show_fixed_bounds       (true      )
                , show_fixed_origin       (false     )
                , show_floating_axis      (false     )
                , show_floating_bounds    (false     )
                , show_ui_message         (true      )
                , show_fulcrum            (false     )
                , show_effects_generation (true      )
                , show_background_bitmap  (true      )
                , loop_animation          (true      )
                , approximate_frame_rate  (true      )
                , wave_show_points        (false     )
                , wave_show_blanks        (false     )
                , wave_real_time          (true      )
                , wave_show_inverted      (false     )
                , wave_clean_screen       (true      )
                , wave_loop               (false     )
                , process_exception       (false     )
{
    if(AIL == NULL)
    {
        AIL = this;
           
        if(SDL_Init(SDL_INIT_VIDEO) < 0)
            SDL_bailout(2);
        if((screen = SDL_SetVideoMode( xres,
                                       yres,
                                       0,
                                         SDL_HWSURFACE
                                       | SDL_ANYFORMAT
                                     )
           ) == NULL
          )
            SDL_bailout(2);
            
        SDL_EnableUNICODE(1);
        SDL_EnableKeyRepeat(250, 1);       

        load_cfg_file();
        load_effects ();
         
        font_size_factor = screen->h / 480;
        if(font_size_factor == 0)
            font_size_factor = 1;
            
        SDL_EventState(SDL_MOUSEMOTION, SDL_IGNORE);
        SDL_EventState(SDL_PRESSED    , SDL_IGNORE);
        SDL_EventState(SDL_RELEASED   , SDL_IGNORE);
        SDL_EventState(SDL_SYSWMEVENT , SDL_IGNORE);
        
        SDL_WM_SetCaption("LaserBoy Multiplatform by James Lehman : Extra Stimulus Inc.", NULL);
        
        if(SDL_LockSurface(screen) < 0)
            SDL_bailout(2);
            
        bmp_clone_SDL_Surface(&bmp, screen, 0);
        bmp_copy(&background    , &bmp);
        bmp_copy(&background_bmp, &bmp);
        recolor_background();
        bmp_to_SDL_Surface(&bmp, screen);
    }
}

//############################################################################
ez_uail_gui::~ez_uail_gui()
{
    bmp_free        (&bmp);
    bmp_free        (&background);
    SDL_FreeSurface (screen);
    SDL_Quit        ();
    AIL = NULL;
}

//############################################################################
void ez_uail_gui::load_cfg_file()
{
    ifstream in(LASERBOY_CFG_FILE, ios::in);
    if(in.is_open())
    {
        int     c;
        string  version;
        //--------------------------------------------------------------------
        in >> version;
        if(version == LASERBOY_CFG_VERSION)
        {
            in >> bg_file_name            ;
            in >> c                       ;
            bg_color.r = c                ;
            in >> c                       ;
            bg_color.g = c                ;
            in >> c                       ;
            bg_color.b = c                ;
            in >> sample_rate             ;
            in >> lit_dwell_overhang      ;
            in >> dwell_on_dot            ;
            in >> font_size_factor        ;
            in >> move_points_per_key     ;
            in >> output_bmp_size         ;
            in >> signal_if_not_z         ;
            in >> fulcrum.x               ;
            in >> fulcrum.y               ;
            in >> fulcrum.z               ;
            in >> view_angle.x            ;
            in >> view_angle.y            ;
            in >> view_angle.z            ;
            in >> view_scale.x            ;
            in >> view_scale.y            ;
            in >> view_offset.x           ;
            in >> view_offset.y           ;
            in >> lit_distance_delta_max  ;
            in >> blank_distance_delta_max;
            in >> longest_dwell_microsec  ;
            in >> insignificant_distance  ;
            in >> zero_angle_tolerance    ;
            in >> recolor_span_factor     ;
            in >> rendered_arc_angle      ;
            in >> rotate_radians_per_key  ;
            in >> scale_per_key           ;
            scale_up_per_key = 1.00 + scale_per_key;
            scale_dn_per_key = 1.00 - scale_per_key;
            in >> frames_per_second       ;
            in >> still_frame_duration    ;
            in >> show_points             ;
            in >> show_blanking           ;
            in >> show_cursors            ;
            in >> show_palettes           ;
            in >> show_stats              ;
            in >> show_fixed_bounds       ;
            in >> show_fixed_origin       ;
            in >> show_floating_axis      ;
            in >> show_floating_bounds    ;
            in >> show_effects_generation ;
            in >> auto_scale_dxf          ;
            in >> maintain_dxf_origin     ;
            in >> loop_animation          ;
            in >> approximate_frame_rate  ;
            in >> black_dwell_points      ;
            in >> auto_flatten_z          ;
            in >> no_equivalent_vectors   ;
            in >> no_dots                 ;
            in >> invert_wave_output      ;
            in >> auto_optimize           ;
            in >> no_black_match          ;
            in >> allow_lit_black         ;
            in >> dumb_file_format        ;         
            in >> wave_show_points        ;
            in >> wave_show_blanks        ;
            in >> wave_real_time          ;
            in >> wave_show_inverted      ;
            in >> wave_clean_screen       ;
            in >> wave_loop               ;
            in >> wave_offsets[0]         ;
            in >> wave_offsets[1]         ;
            in >> wave_offsets[2]         ;
            in >> wave_offsets[3]         ;
            in >> wave_offsets[4]         ;
            in >> wave_offsets[5]         ;
            in >> wave_offsets[6]         ;
            in >> wave_offsets[7]         ;
            in >> c                       ;
            black_threshold = c           ;
            in >> signal_bit_mask[0]      ;
            in >> signal_bit_mask[1]      ;
            in >> signal_bit_mask[2]      ;
            in >> signal_bit_mask[3]      ;
            in >> signal_bit_mask[4]      ;
            in >> signal_bit_mask[5]      ;
            in >> signal_bit_mask[6]      ;
            in >> signal_bit_mask[7]      ;
        }
        //--------------------------------------------------------------------
        else if(version == "LaserBoy.wtf-01-16-2009")
        {
            in >> bg_file_name            ;
            in >> c                       ;
            bg_color.r = c                ;
            in >> c                       ;
            bg_color.g = c                ;
            in >> c                       ;
            bg_color.b = c                ;
            in >> sample_rate             ;
            in >> lit_dwell_overhang      ;
            in >> dwell_on_dot            ;
            in >> font_size_factor        ;
            in >> move_points_per_key     ;
            in >> output_bmp_size         ;
            in >> fulcrum.x               ;
            in >> fulcrum.y               ;
            in >> fulcrum.z               ;
            in >> view_angle.x            ;
            in >> view_angle.y            ;
            in >> view_angle.z            ;
            in >> view_scale.x            ;
            in >> view_scale.y            ;
            in >> view_offset.x           ;
            in >> view_offset.y           ;
            in >> lit_distance_delta_max  ;
            in >> blank_distance_delta_max;
            in >> longest_dwell_microsec  ;
            in >> insignificant_distance  ;
            in >> zero_angle_tolerance    ;
            in >> recolor_span_factor     ;
            in >> rendered_arc_angle      ;
            in >> rotate_radians_per_key  ;
            in >> scale_per_key           ;
            scale_up_per_key = 1.00 + scale_per_key;
            scale_dn_per_key = 1.00 - scale_per_key;
            in >> frames_per_second       ;
            in >> still_frame_duration    ;
            in >> show_points             ;
            in >> show_blanking           ;
            in >> show_cursors            ;
            in >> show_palettes           ;
            in >> show_stats              ;
            in >> show_fixed_bounds       ;
            in >> show_fixed_origin       ;
            in >> show_floating_axis      ;
            in >> show_floating_bounds    ;
            in >> show_effects_generation ;
            in >> auto_scale_dxf          ;
            in >> maintain_dxf_origin     ;
            in >> loop_animation          ;
            in >> approximate_frame_rate  ;
            in >> black_dwell_points      ;
            in >> auto_flatten_z          ;
            in >> no_equivalent_vectors   ;
            in >> no_dots                 ;
            in >> invert_wave_output      ;
            in >> auto_optimize           ;
            in >> no_black_match          ;
            in >> allow_lit_black         ;
            in >> dumb_file_format        ;         
            in >> wave_show_points        ;
            in >> wave_show_blanks        ;
            in >> wave_real_time          ;
            in >> wave_show_inverted      ;
            in >> wave_clean_screen       ;
            in >> wave_loop               ;
            in >> wave_offsets[0]         ;
            in >> wave_offsets[1]         ;
            in >> wave_offsets[2]         ;
            in >> wave_offsets[3]         ;
            in >> wave_offsets[4]         ;
            in >> wave_offsets[5]         ;
            in >> c                       ;
            black_threshold = c           ;
        }
        //--------------------------------------------------------------------
        else if(version == "LaserBoy.wtf-01-09-2009")
        {
            in >> bg_file_name            ;
            in >> c                       ;
            bg_color.r = c                ;
            in >> c                       ;
            bg_color.g = c                ;
            in >> c                       ;
            bg_color.b = c                ;
            in >> sample_rate             ;
            in >> lit_dwell_overhang      ;
            in >> font_size_factor        ;
            in >> move_points_per_key     ;
            in >> output_bmp_size         ;
            in >> fulcrum.x               ;
            in >> fulcrum.y               ;
            in >> fulcrum.z               ;
            in >> view_angle.x            ;
            in >> view_angle.y            ;
            in >> view_angle.z            ;
            in >> view_scale.x            ;
            in >> view_scale.y            ;
            in >> view_offset.x           ;
            in >> view_offset.y           ;
            in >> lit_distance_delta_max  ;
            in >> blank_distance_delta_max;
            in >> longest_dwell_microsec  ;
            in >> insignificant_distance  ;
            in >> zero_angle_tolerance    ;
            in >> recolor_span_factor     ;
            in >> rendered_arc_angle      ;
            in >> rotate_radians_per_key  ;
            in >> frames_per_second       ;
            in >> still_frame_duration    ;
            in >> show_points             ;
            in >> show_blanking           ;
            in >> show_cursors            ;
            in >> show_palettes           ;
            in >> show_stats              ;
            in >> show_fixed_bounds       ;
            in >> show_fixed_origin       ;
            in >> show_floating_axis      ;
            in >> show_floating_bounds    ;
            in >> show_effects_generation ;
            in >> auto_scale_dxf          ;
            in >> maintain_dxf_origin     ;
            in >> loop_animation          ;
            in >> approximate_frame_rate  ;
            in >> black_dwell_points      ;
            in >> auto_flatten_z          ;
            in >> no_equivalent_vectors   ;
            in >> no_dots                 ;
            in >> invert_wave_output      ;
            in >> auto_optimize           ;
            in >> no_black_match          ;
            in >> allow_lit_black         ;
            in >> dumb_file_format        ;         
            in >> wave_show_points        ;
            in >> wave_show_blanks        ;
            in >> wave_real_time          ;
            in >> wave_show_inverted      ;
            in >> wave_clean_screen       ;
            in >> wave_loop               ;
            in >> wave_offsets[0]         ;
            in >> wave_offsets[1]         ;
            in >> wave_offsets[2]         ;
            in >> wave_offsets[3]         ;
            in >> wave_offsets[4]         ;
            in >> wave_offsets[5]         ;
            in >> c                       ;
            black_threshold = c           ;
        }        
        //--------------------------------------------------------------------
        else if(version == "LaserBoy.wtf-01-06-2009")
        {
            in >> bg_file_name            ;
            in >> c                       ;
            bg_color.r = c                ;
            in >> c                       ;
            bg_color.g = c                ;
            in >> c                       ;
            bg_color.b = c                ;
            in >> sample_rate             ;
            in >> lit_dwell_overhang      ;
            in >> font_size_factor        ;
            in >> move_points_per_key     ;
            in >> output_bmp_size         ;
            in >> fulcrum.x               ;
            in >> fulcrum.y               ;
            in >> fulcrum.z               ;
            in >> view_angle.x            ;
            in >> view_angle.y            ;
            in >> view_angle.z            ;
            in >> view_scale.x            ;
            in >> view_scale.y            ;
            in >> view_offset.x           ;
            in >> view_offset.y           ;
            in >> lit_distance_delta_max  ;
            in >> blank_distance_delta_max;
            in >> longest_dwell_microsec  ;
            in >> insignificant_distance  ;
            in >> zero_angle_tolerance    ;
            in >> recolor_span_factor     ;
            in >> rendered_arc_angle      ;
            in >> rotate_radians_per_key  ;
            in >> frames_per_second       ;
            in >> still_frame_duration    ;
            in >> show_points             ;
            in >> show_blanking           ;
            in >> show_cursors            ;
            in >> show_palettes           ;
            in >> show_stats              ;
            in >> show_fixed_bounds       ;
            in >> show_fixed_origin       ;
            in >> show_floating_axis      ;
            in >> show_floating_bounds    ;
            in >> show_effects_generation ;
            in >> auto_scale_dxf          ;
            in >> maintain_dxf_origin     ;
            in >> loop_animation          ;
            in >> approximate_frame_rate  ;
            in >> black_dwell_points      ;
            in >> auto_flatten_z          ;
            in >> no_equivalent_vectors   ;
            in >> invert_wave_output      ;
            in >> auto_optimize           ;
            in >> no_black_match          ;
            in >> allow_lit_black         ;
            in >> dumb_file_format        ;         
            in >> wave_show_points        ;
            in >> wave_show_blanks        ;
            in >> wave_real_time          ;
            in >> wave_show_inverted      ;
            in >> wave_clean_screen       ;
            in >> wave_loop               ;
            in >> wave_offsets[0]         ;
            in >> wave_offsets[1]         ;
            in >> wave_offsets[2]         ;
            in >> wave_offsets[3]         ;
            in >> wave_offsets[4]         ;
            in >> wave_offsets[5]         ;
            in >> c                       ;
            black_threshold = c           ;
        }
        //--------------------------------------------------------------------
        else if(version == "LaserBoy.wtf-11-11-2008")
        {
            in >> bg_file_name            ;
            in >> c                       ;
            bg_color.r = c                ;
            in >> c                       ;
            bg_color.g = c                ;
            in >> c                       ;
            bg_color.b = c                ;
            in >> sample_rate             ;
            in >> lit_dwell_overhang      ;
            in >> font_size_factor        ;
            in >> move_points_per_key     ;
            in >> output_bmp_size         ;
            in >> fulcrum.x               ;
            in >> fulcrum.y               ;
            in >> fulcrum.z               ;
            in >> view_angle.x            ;
            in >> view_angle.y            ;
            in >> view_angle.z            ;
            in >> view_scale.x            ;
            in >> view_scale.y            ;
            in >> view_offset.x           ;
            in >> view_offset.y           ;
            in >> lit_distance_delta_max  ;
            in >> blank_distance_delta_max;
            in >> longest_dwell_microsec  ;
            in >> insignificant_distance  ;
            in >> zero_angle_tolerance    ;
            in >> recolor_span_factor     ;
            in >> rendered_arc_angle      ;
            in >> rotate_radians_per_key  ;
            in >> frames_per_second       ;
            in >> still_frame_duration    ;
            in >> show_points             ;
            in >> show_blanking           ;
            in >> show_cursors            ;
            in >> show_palettes           ;
            in >> show_stats              ;
            in >> show_fixed_bounds       ;
            in >> show_fixed_origin       ;
            in >> show_floating_axis      ;
            in >> show_floating_bounds    ;
            in >> show_effects_generation ;
            in >> auto_scale_dxf          ;
            in >> maintain_dxf_origin     ;
            in >> loop_animation          ;
            in >> approximate_frame_rate  ;
            in >> black_dwell_points      ;
            in >> auto_flatten_z          ;
            in >> no_equivalent_vectors   ;
            in >> invert_wave_output      ;
            in >> auto_optimize           ;
            in >> no_black_match          ;            
            in >> dumb_file_format        ;         
            in >> wave_show_points        ;
            in >> wave_show_blanks        ;
            in >> wave_real_time          ;
            in >> wave_show_inverted      ;
            in >> wave_clean_screen       ;
            in >> wave_loop               ;
            in >> wave_offsets[0]         ;
            in >> wave_offsets[1]         ;
            in >> wave_offsets[2]         ;
            in >> wave_offsets[3]         ;
            in >> wave_offsets[4]         ;
            in >> wave_offsets[5]         ;
            in >> c                       ;
            black_threshold = c           ;
        }
        //--------------------------------------------------------------------
        else if(version == "LaserBoy.wtf-10-10-2008")
        {
            in >> bg_file_name           ;
            in >> c                      ;
            bg_color.r = c               ;
            in >> c                      ;
            bg_color.g = c               ;
            in >> c                      ;
            bg_color.b = c               ;
            in >> sample_rate            ;
            in >> lit_dwell_overhang     ;
            in >> font_size_factor       ;
            in >> move_points_per_key    ;
            in >> output_bmp_size        ;
            in >> fulcrum.x              ;
            in >> fulcrum.y              ;
            in >> fulcrum.z              ;
            in >> view_angle.x           ;
            in >> view_angle.y           ;
            in >> view_angle.z           ;
            in >> view_scale.x           ;
            in >> view_scale.y           ;
            in >> view_offset.x          ;
            in >> view_offset.y          ;
            in >> lit_distance_delta_max ;
            blank_distance_delta_max = lit_distance_delta_max;
            in >> longest_dwell_microsec ;
            in >> insignificant_distance ;
            in >> zero_angle_tolerance   ;
            in >> recolor_span_factor    ;
            in >> rendered_arc_angle     ;
            in >> rotate_radians_per_key ;
            in >> frames_per_second      ;
            in >> still_frame_duration   ;
            in >> show_points            ;
            in >> show_blanking          ;
            in >> show_cursors           ;
            in >> show_palettes          ;
            in >> show_stats             ;
            in >> show_fixed_bounds      ;
            in >> show_fixed_origin      ;
            in >> show_floating_axis     ;
            in >> show_floating_bounds   ;
            in >> show_effects_generation;
            in >> auto_scale_dxf         ;
            in >> maintain_dxf_origin    ;
            in >> loop_animation         ;
            in >> approximate_frame_rate ;
            in >> black_dwell_points     ;
            in >> auto_flatten_z         ;
            in >> no_equivalent_vectors  ;
            in >> invert_wave_output     ;
            in >> auto_optimize          ;
            in >> dumb_file_format       ;         
            in >> wave_show_points       ;
            in >> wave_show_blanks       ;
            in >> wave_real_time         ;
            in >> wave_show_inverted     ;
            in >> wave_clean_screen      ;
            in >> wave_loop              ;
            in >> wave_offsets[0]        ;
            in >> wave_offsets[1]        ;
            in >> wave_offsets[2]        ;
            in >> wave_offsets[3]        ;
            in >> wave_offsets[4]        ;
            in >> wave_offsets[5]        ;
        }
        //--------------------------------------------------------------------
        else if(version == "LaserBoy.wtf-09-03-2008")
        {
            in >> bg_file_name           ;
            in >> c                      ;
            bg_color.r = c               ;
            in >> c                      ;
            bg_color.g = c               ;
            in >> c                      ;
            bg_color.b = c               ;
            in >> sample_rate            ;
            in >> lit_dwell_overhang     ;
            in >> font_size_factor       ;
            in >> move_points_per_key    ;
            in >> output_bmp_size        ;
            in >> fulcrum.x              ;
            in >> fulcrum.y              ;
            in >> fulcrum.z              ;
            in >> view_angle.x           ;
            in >> view_angle.y           ;
            in >> view_angle.z           ;
            in >> view_scale.x           ;
            in >> view_scale.y           ;
            in >> view_offset.x          ;
            in >> view_offset.y          ;
            in >> lit_distance_delta_max ;
            blank_distance_delta_max = lit_distance_delta_max;
            in >> longest_dwell_microsec ;
            in >> insignificant_distance ;
            in >> zero_angle_tolerance   ;
            in >> recolor_span_factor    ;
            in >> rendered_arc_angle     ;
            in >> rotate_radians_per_key ;
            in >> frames_per_second      ;
            in >> still_frame_duration   ;
            in >> show_points            ;
            in >> show_blanking          ;
            in >> show_cursors           ;
            in >> show_palettes          ;
            in >> show_stats             ;
            in >> show_fixed_bounds      ;
            in >> show_fixed_origin      ;
            in >> show_floating_axis     ;
            in >> show_floating_bounds   ;
            in >> show_effects_generation;
            in >> auto_scale_dxf         ;
            in >> maintain_dxf_origin    ;
            in >> loop_animation         ;
            in >> approximate_frame_rate ;
            in >> black_dwell_points     ;
            in >> auto_flatten_z         ;
            in >> no_equivalent_vectors  ;
            in >> invert_wave_output     ;
            in >> auto_optimize          ;
            in >> dumb_file_format       ;         
            in >> wave_show_points       ;
            in >> wave_show_blanks       ;
            in >> wave_real_time         ;
            in >> wave_show_inverted     ;
            in >> wave_clean_screen      ;
            in >> wave_offsets[0]        ;
            in >> wave_offsets[1]        ;
            in >> wave_offsets[2]        ;
            in >> wave_offsets[3]        ;
            in >> wave_offsets[4]        ;
            in >> wave_offsets[5]        ;
        }
        //--------------------------------------------------------------------
        else if(version == "LaserBoy.wtf-08-28-2008")
        {
            in >> bg_file_name           ;
            in >> c                      ;
            bg_color.r = c               ;
            in >> c                      ;
            bg_color.g = c               ;
            in >> c                      ;
            bg_color.b = c               ;
            in >> sample_rate            ;
            in >> lit_dwell_overhang     ;
            in >> font_size_factor       ;
            in >> move_points_per_key    ;
            in >> output_bmp_size        ;
            in >> fulcrum.x              ;
            in >> fulcrum.y              ;
            in >> fulcrum.z              ;
            in >> view_angle.x           ;
            in >> view_angle.y           ;
            in >> view_angle.z           ;
            in >> view_scale.x           ;
            in >> view_scale.y           ;
            in >> view_offset.x          ;
            in >> view_offset.y          ;
            in >> lit_distance_delta_max ;
            blank_distance_delta_max = lit_distance_delta_max;
            in >> longest_dwell_microsec ;
            in >> insignificant_distance ;
            in >> zero_angle_tolerance   ;
            in >> recolor_span_factor    ;
            in >> rendered_arc_angle     ;
            in >> rotate_radians_per_key ;
            in >> frames_per_second      ;
            in >> still_frame_duration   ;
            in >> show_points            ;
            in >> show_blanking          ;
            in >> show_cursors           ;
            in >> show_palettes          ;
            in >> show_stats             ;
            in >> show_fixed_bounds      ;
            in >> show_fixed_origin      ;
            in >> show_floating_axis     ;
            in >> show_floating_bounds   ;
            in >> show_effects_generation;
            in >> auto_scale_dxf         ;
            in >> maintain_dxf_origin    ;
            in >> loop_animation         ;
            in >> approximate_frame_rate ;
            in >> black_dwell_points     ;
            in >> auto_flatten_z         ;
            in >> no_equivalent_vectors  ;
            in >> invert_wave_output     ;
            in >> auto_optimize          ;
            in >> wave_show_points       ;
            in >> wave_show_blanks       ;
            in >> wave_real_time         ;
            in >> wave_show_inverted     ;
            in >> wave_clean_screen      ;
            in >> wave_offsets[0]        ;
            in >> wave_offsets[1]        ;
            in >> wave_offsets[2]        ;
            in >> wave_offsets[3]        ;
            in >> wave_offsets[4]        ;
            in >> wave_offsets[5]        ;
        }
        //--------------------------------------------------------------------
        else if(version == "LaserBoy.wtf-07-24-2008")
        {
            in >> bg_file_name           ;
            in >> c                      ;
            bg_color.r = c               ;
            in >> c                      ;
            bg_color.g = c               ;
            in >> c                      ;
            bg_color.b = c               ;
            in >> sample_rate            ;
            in >> lit_dwell_overhang     ;
            in >> font_size_factor       ;
            in >> move_points_per_key    ;
            in >> output_bmp_size        ;
            in >> fulcrum.x              ;
            in >> fulcrum.y              ;
            in >> fulcrum.z              ;
            in >> view_angle.x           ;
            in >> view_angle.y           ;
            in >> view_angle.z           ;
            in >> view_scale.x           ;
            in >> view_scale.y           ;
            in >> view_scale.z           ;
            in >> view_offset.x          ;
            in >> view_offset.y          ;
            in >> view_offset.z          ;
            in >> lit_distance_delta_max ;
            blank_distance_delta_max = lit_distance_delta_max;
            in >> longest_dwell_microsec ;
            in >> insignificant_distance ;
            in >> zero_angle_tolerance   ;
            in >> recolor_span_factor    ;
            in >> rendered_arc_angle     ;
            in >> rotate_radians_per_key ;
            in >> frames_per_second      ;
            in >> still_frame_duration   ;
            in >> show_points            ;
            in >> show_blanking          ;
            in >> show_cursors           ;
            in >> show_palettes          ;
            in >> show_stats             ;
            in >> show_fixed_bounds      ;
            in >> show_fixed_origin      ;
            in >> show_floating_axis     ;
            in >> show_floating_bounds   ;
            in >> show_effects_generation;
            in >> auto_scale_dxf         ;
            in >> maintain_dxf_origin    ;
            in >> loop_animation         ;
            in >> approximate_frame_rate ;
            in >> black_dwell_points     ;
            in >> auto_flatten_z         ;
            in >> no_equivalent_vectors  ;
            in >> invert_wave_output     ;
            in >> auto_optimize          ;
            in >> wave_show_points       ;
            in >> wave_show_blanks       ;
            in >> wave_real_time         ;
            in >> wave_show_inverted     ;
            in >> wave_clean_screen      ;
            in >> wave_offsets[0]        ;
            in >> wave_offsets[1]        ;
            in >> wave_offsets[2]        ;
            in >> wave_offsets[3]        ;
            in >> wave_offsets[4]        ;
            in >> wave_offsets[5]        ;
        }
        //--------------------------------------------------------------------
        else if(version == "LaserBoy.wtf-07-17-2008")
        {
            in >> bg_file_name           ;
            in >> c                      ;
            bg_color.r = c               ;
            in >> c                      ;
            bg_color.g = c               ;
            in >> c                      ;
            bg_color.b = c               ;
            in >> sample_rate            ;
            in >> lit_dwell_overhang     ;
            in >> font_size_factor       ;
            in >> move_points_per_key    ;
            in >> output_bmp_size        ;
            in >> fulcrum.x              ;
            in >> fulcrum.y              ;
            in >> fulcrum.z              ;
            in >> view_angle.x           ;
            in >> view_angle.y           ;
            in >> view_angle.z           ;
            in >> view_scale.x           ;
            in >> view_scale.y           ;
            in >> view_scale.z           ;
            in >> view_offset.x          ;
            in >> view_offset.y          ;
            in >> view_offset.z          ;
            in >> lit_distance_delta_max ;
            blank_distance_delta_max = lit_distance_delta_max;
            in >> longest_dwell_microsec ;
            in >> insignificant_distance ;
            in >> zero_angle_tolerance   ;
            in >> recolor_span_factor    ;
            in >> rendered_arc_angle     ;
            in >> frames_per_second      ;
            in >> still_frame_duration   ;
            in >> show_points            ;
            in >> show_blanking          ;
            in >> show_cursors           ;
            in >> show_palettes          ;
            in >> show_stats             ;
            in >> show_fixed_bounds      ;
            in >> show_fixed_origin      ;
            in >> show_floating_axis     ;
            in >> show_floating_bounds   ;
            in >> show_effects_generation;
            in >> auto_scale_dxf         ;
            in >> maintain_dxf_origin    ;
            in >> loop_animation         ;
            in >> approximate_frame_rate ;
            in >> black_dwell_points     ;
            in >> auto_flatten_z         ;
            in >> no_equivalent_vectors  ;
            in >> invert_wave_output     ;
            in >> auto_optimize          ;
            in >> wave_show_points       ;
            in >> wave_show_blanks       ;
            in >> wave_real_time         ;
            in >> wave_show_inverted     ;
            in >> wave_clean_screen      ;
            in >> wave_offsets[0]        ;
            in >> wave_offsets[1]        ;
            in >> wave_offsets[2]        ;
            in >> wave_offsets[3]        ;
            in >> wave_offsets[4]        ;
            in >> wave_offsets[5]        ;
        }
        //--------------------------------------------------------------------        
        else if(version == "LaserBoy.wtf-07-15-2008")
        {
            in >> bg_file_name           ;
            in >> c                      ;
            bg_color.r = c               ;
            in >> c                      ;
            bg_color.g = c               ;
            in >> c                      ;
            bg_color.b = c               ;
            in >> sample_rate            ;
            in >> lit_dwell_overhang     ;
            in >> font_size_factor       ;
            in >> move_points_per_key    ;
            in >> output_bmp_size        ;
            in >> fulcrum.x              ;
            in >> fulcrum.y              ;
            in >> fulcrum.z              ;
            in >> view_angle.x           ;
            in >> view_angle.y           ;
            in >> view_angle.z           ;
            in >> view_scale.x           ;
            in >> view_scale.y           ;
            in >> view_scale.z           ;
            in >> view_offset.x          ;
            in >> view_offset.y          ;
            in >> view_offset.z          ;
            in >> lit_distance_delta_max ;
            blank_distance_delta_max = lit_distance_delta_max;
            in >> longest_dwell_microsec ;
            in >> insignificant_distance ;
            in >> zero_angle_tolerance   ;
            in >> recolor_span_factor    ;
            in >> rendered_arc_angle     ;
            in >> frames_per_second      ;
            in >> still_frame_duration   ;
            in >> show_points            ;
            in >> show_blanking          ;
            in >> show_cursors           ;
            in >> show_palettes          ;
            in >> show_stats             ;
            in >> show_fixed_bounds      ;
            in >> show_fixed_origin      ;
            in >> show_floating_axis     ;
            in >> show_floating_bounds   ;
            in >> show_effects_generation;
            in >> auto_scale_dxf         ;
            in >> maintain_dxf_origin    ;
            in >> loop_animation         ;
            in >> approximate_frame_rate ;
            in >> black_dwell_points     ;
            in >> auto_flatten_z         ;
            in >> no_equivalent_vectors  ;
            in >> auto_optimize          ;
            in >> wave_show_points       ;
            in >> wave_show_blanks       ;
            in >> wave_real_time         ;
            in >> wave_show_inverted     ;
            in >> wave_clean_screen      ;
            in >> wave_offsets[0]        ;
            in >> wave_offsets[1]        ;
            in >> wave_offsets[2]        ;
            in >> wave_offsets[3]        ;
            in >> wave_offsets[4]        ;
            in >> wave_offsets[5]        ;
        }        
        //--------------------------------------------------------------------
        else if(version == "LaserBoy.wtf-06-03-2008")
        {
            in >> bg_file_name           ;
            in >> c                      ;
            bg_color.r = c               ;
            in >> c                      ;
            bg_color.g = c               ;
            in >> c                      ;
            bg_color.b = c               ;
            in >> sample_rate            ;
            in >> lit_dwell_overhang     ;
            in >> font_size_factor       ;
            in >> move_points_per_key    ;
            in >> output_bmp_size        ;
            in >> fulcrum.x              ;
            in >> fulcrum.y              ;
            in >> fulcrum.z              ;
            in >> view_angle.x           ;
            in >> view_angle.y           ;
            in >> view_angle.z           ;
            in >> view_scale.x           ;
            in >> view_scale.y           ;
            in >> view_scale.z           ;
            in >> view_offset.x          ;
            in >> view_offset.y          ;
            in >> view_offset.z          ;
            in >> lit_distance_delta_max ;
            blank_distance_delta_max = lit_distance_delta_max;
            in >> longest_dwell_microsec ;
            in >> insignificant_distance ;
            in >> zero_angle_tolerance   ;
            in >> recolor_span_factor    ;
            in >> rendered_arc_angle     ;
            in >> frames_per_second      ;
            in >> still_frame_duration   ;
            in >> show_points            ;
            in >> show_blanking          ;
            in >> show_cursors           ;
            in >> show_palettes          ;
            in >> show_stats             ;
            in >> show_fixed_bounds      ;
            in >> show_fixed_origin      ;
            in >> show_floating_axis     ;
            in >> show_floating_bounds   ;
            in >> show_effects_generation;
            in >> auto_scale_dxf         ;
            in >> maintain_dxf_origin    ;
            in >> loop_animation         ;
            in >> approximate_frame_rate ;
            in >> black_dwell_points     ;
            in >> auto_flatten_z         ;
            in >> auto_optimize          ;
            in >> wave_show_points       ;
            in >> wave_show_blanks       ;
            in >> wave_real_time         ;
            in >> wave_show_inverted     ;
            in >> wave_clean_screen      ;
            in >> wave_offsets[0]        ;
            in >> wave_offsets[1]        ;
            in >> wave_offsets[2]        ;
            in >> wave_offsets[3]        ;
            in >> wave_offsets[4]        ;
            in >> wave_offsets[5]        ;
        }
        in.close();
        //--------------------------------------------------------------------
        if(version != LASERBOY_CFG_VERSION)        
            save_cfg_file();
    }
    return;
}

//############################################################################
void ez_uail_gui::save_cfg_file()
{
    ofstream out(LASERBOY_CFG_FILE, ios::out);
    if(out.is_open())
    {
        out << LASERBOY_CFG_VERSION     << endl;
        out << bg_file_name.c_str()     << endl;
        out << (int)(bg_color.r)        << endl;
        out << (int)(bg_color.g)        << endl;
        out << (int)(bg_color.b)        << endl;
        out << sample_rate              << endl;
        out << lit_dwell_overhang       << endl;
        out << dwell_on_dot             << endl;
        out << font_size_factor         << endl;
        out << move_points_per_key      << endl;
        out << output_bmp_size          << endl;
        out << signal_if_not_z          << endl;        
        out << fulcrum.x                << endl;
        out << fulcrum.y                << endl;
        out << fulcrum.z                << endl;
        out << view_angle.x             << endl;
        out << view_angle.y             << endl;
        out << view_angle.z             << endl;
        out << view_scale.x             << endl;
        out << view_scale.y             << endl;
        out << view_offset.x            << endl;
        out << view_offset.y            << endl;
        out << lit_distance_delta_max   << endl;
        out << blank_distance_delta_max << endl;
        out << longest_dwell_microsec   << endl;
        out << insignificant_distance   << endl;
        out << zero_angle_tolerance     << endl;
        out << recolor_span_factor      << endl;
        out << rendered_arc_angle       << endl;
        out << rotate_radians_per_key   << endl;
        out << scale_per_key            << endl;
        out << frames_per_second        << endl;
        out << still_frame_duration     << endl;
        out << show_points              << endl;
        out << show_blanking            << endl;
        out << show_cursors             << endl;
        out << show_palettes            << endl;
        out << show_stats               << endl;
        out << show_fixed_bounds        << endl;
        out << show_fixed_origin        << endl;
        out << show_floating_axis       << endl;
        out << show_floating_bounds     << endl;
        out << show_effects_generation  << endl;
        out << auto_scale_dxf           << endl;
        out << maintain_dxf_origin      << endl;
        out << loop_animation           << endl;
        out << approximate_frame_rate   << endl;
        out << black_dwell_points       << endl;
        out << auto_flatten_z           << endl;
        out << no_equivalent_vectors    << endl;
        out << no_dots                  << endl;
        out << invert_wave_output       << endl;        
        out << auto_optimize            << endl;
        out << no_black_match           << endl;
        out << allow_lit_black          << endl;
        out << dumb_file_format         << endl;        
        out << wave_show_points         << endl;
        out << wave_show_blanks         << endl;
        out << wave_real_time           << endl;
        out << wave_show_inverted       << endl;
        out << wave_clean_screen        << endl;    
        out << wave_loop                << endl;    
        out << wave_offsets[0]          << endl;
        out << wave_offsets[1]          << endl;
        out << wave_offsets[2]          << endl;
        out << wave_offsets[3]          << endl;
        out << wave_offsets[4]          << endl;
        out << wave_offsets[5]          << endl;
        out << wave_offsets[6]          << endl;
        out << wave_offsets[7]          << endl;
        out << (int)(black_threshold)   << endl;
        out << signal_bit_mask[0]       << endl;
        out << signal_bit_mask[1]       << endl;
        out << signal_bit_mask[2]       << endl;
        out << signal_bit_mask[3]       << endl;
        out << signal_bit_mask[4]       << endl;
        out << signal_bit_mask[5]       << endl;
        out << signal_bit_mask[6]       << endl;
        out << signal_bit_mask[7]       << endl << endl;
        out << "If this file is currupted, delete it!" << endl;
        out.close();
    }
    return;
}

//############################################################################
void ez_uail_gui::display(string ui_message)
{
    if(palette_index != current_frame().palette_index)
        palette_index = current_frame().palette_index;
    current_frame().render_in_ezfb_bitmap(ui_message);
    bmp_to_SDL_Surface(&bmp, screen);
    return;
}

//############################################################################
string ez_uail_gui::display_prompt_and_echo_name(string prompt)
{
    bool   done         = false;
    int    x_off        = EZ_STATS_DISPLAY_OFFSET * font_size_factor,
           y_off        = BMP.yres - 24 * font_size_factor,
           x_gap        = 8 * font_size_factor,
           y_line       = x_gap;
    u_int  bmp_bg_color = BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b);  
    string input;    
    
    process_exception = false;
        
    bmp_printf(   &BMP, x_off, y_off, BMP.white, bmp_bg_color, 0, font_size_factor
                , prompt.c_str()
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, BMP.make_rgb(0,255,255), bmp_bg_color, 0, font_size_factor
                , "_"
              );
    bmp_to_SDL_Surface(&bmp, screen);

    while(!done)
    {
        do
        {
            SDL_WaitEvent(&event);
        } while(event.type != SDL_KEYDOWN);
        if(event.key.keysym.unicode == SDLK_ESCAPE)
        {
            process_exception = true;
            return "";
        }
        if(event.key.keysym.unicode == SDLK_RETURN)
            done = true;
        else
        {
            switch(event.key.keysym.unicode)
            {
                //-------------------------------------------------    
                case SDLK_DELETE:
                case SDLK_BACKSPACE:
                    if(input.size())
                    {
                        input = input.substr(0, input.size() - 1);
                        x_off -= x_gap;
                        bmp_printf(   &BMP, x_off, y_off, BMP.make_rgb(0,255,255), bmp_bg_color, 0, font_size_factor
                                    , " "
                                  );
                        bmp_to_SDL_Surface(&bmp, screen);
                    }                         
                    break;
                //-------------------------------------------------    
                default:
                    if(    isalnum(event.key.keysym.unicode)
                        || event.key.keysym.unicode == '_'
                        || event.key.keysym.unicode == '.'
                      )
                    {
                        bmp_printf(   &BMP, x_off, y_off, BMP.make_rgb(0,255,255), bmp_bg_color, 0, font_size_factor
                                    , "%c"
                                    , event.key.keysym.unicode
                                  );
                        bmp_to_SDL_Surface(&bmp, screen);
                        input += event.key.keysym.unicode;
                        x_off += x_gap;
                    }                            
                    break;                         
                //-------------------------------------------------    
            } // end switch(event.key.keysym.unicode)
        } // end else of if(event.key.keysym.unicode == SDLK_RETURN)
    } // end while(!done)    
    bmp_printf(   &BMP
                , x_off - x_gap * input.size()
                , y_off
                , BMP.make_rgb(255,0,0)
                , bmp_bg_color
                , 0
                , font_size_factor
                , input.c_str()
              );
    bmp_to_SDL_Surface(&bmp, screen);
    return input;
}

//############################################################################
string ez_uail_gui::display_prompt_and_echo_string(string prompt)
{
    bool   done         = false;
    int    x_off        = EZ_STATS_DISPLAY_OFFSET * font_size_factor,
           y_off        = BMP.yres - 24 * font_size_factor,
           x_gap        = 8 * font_size_factor,
           y_line       = 8 * font_size_factor;
    u_int  bmp_bg_color = BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b);
    string input;

    process_exception = false;
    bmp_printf(   &BMP, x_off, y_off, BMP.white, bmp_bg_color, 0, font_size_factor
                , prompt.c_str()
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, BMP.make_rgb(0,255,255), bmp_bg_color, 0, font_size_factor
                , "_"
              );
    bmp_to_SDL_Surface(&bmp, screen);
              
    while(!done)
    {
        do
        {
            SDL_WaitEvent(&event);
        } while(event.type != SDL_KEYDOWN);
        if(event.key.keysym.unicode == SDLK_ESCAPE)
        {
            process_exception = true;
            return "";
        }
        if(event.key.keysym.unicode == SDLK_RETURN)
            done = true;
        else
        {
            switch(event.key.keysym.unicode)
            {
                //-------------------------------------------------    
                case SDLK_DELETE:
                case SDLK_BACKSPACE:
                    if(input.size())
                    {
                        input = input.substr(0, input.size() - 1);
                        x_off -= x_gap;
                        bmp_printf(   &BMP, x_off, y_off, BMP.make_rgb(0,255,255), bmp_bg_color, 0, font_size_factor
                                    , " "
                                  );
                        bmp_to_SDL_Surface(&bmp, screen);
                    }                         
                    break;
                //-------------------------------------------------    
                default:
                    if(    isgraph(event.key.keysym.unicode)
                        || event.key.keysym.unicode == ' '
                      )
                    {
                        bmp_printf(   &BMP, x_off, y_off, BMP.make_rgb(0,255,255), bmp_bg_color, 0, font_size_factor
                                    , "%c"
                                    , event.key.keysym.unicode
                                  );
                        bmp_to_SDL_Surface(&bmp, screen);
                        input += event.key.keysym.unicode;
                        x_off += x_gap;
                    }                            
                    break;                         
                //-------------------------------------------------    
            } // end switch(event.key.keysym.unicode)
        } // end else of if(event.key.keysym.unicode == SDLK_RETURN)
    } // end while(!done)
    bmp_printf(   &BMP
                , x_off - x_gap * input.size()
                , y_off
                , BMP.make_rgb(255,0,0)
                , bmp_bg_color
                , 0
                , font_size_factor
                , input.c_str()
              );
    bmp_to_SDL_Surface(&bmp, screen);
    return input;
}

//############################################################################
float ez_uail_gui::display_prompt_and_echo_float(string prompt)
{
    bool   done         = false;
    int    x_off        = EZ_STATS_DISPLAY_OFFSET * font_size_factor,
           y_off        = BMP.yres - 24 * font_size_factor,
           x_gap        = 8 * font_size_factor,
           y_line       = 8 * font_size_factor;
    u_int  bmp_bg_color = BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b);
    float  result;
    string input;

    process_exception = false;
    bmp_printf(   &BMP, x_off, y_off, BMP.white, bmp_bg_color, 0, font_size_factor
                , prompt.c_str()
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, BMP.make_rgb(0,255,255), bmp_bg_color, 0, font_size_factor
                , "_"
              );
    bmp_to_SDL_Surface(&bmp, screen);

    while(!done)
    {
        do
        {
            SDL_WaitEvent(&event);
        } while(event.type != SDL_KEYDOWN);
        if(event.key.keysym.unicode == SDLK_ESCAPE)
        {
            process_exception = true;
            return 0.0;
        }
        if(event.key.keysym.unicode == SDLK_RETURN)
            done = true;
        else
        {
            switch(event.key.keysym.unicode)
            {
                //-------------------------------------------------    
                case SDLK_DELETE:
                case SDLK_BACKSPACE:
                    if(input.size())
                    {
                        input = input.substr(0, input.size() - 1);
                        x_off -= x_gap;
                        bmp_printf(   &BMP, x_off, y_off, BMP.make_rgb(0,255,255), bmp_bg_color, 0, font_size_factor
                                    , " "
                                  );
                        bmp_to_SDL_Surface(&bmp, screen);
                    }                         
                    break;
                //-------------------------------------------------    
                default:
                    if(    isdigit(event.key.keysym.unicode)
                        || (    event.key.keysym.unicode == '.'
                             && input.find('.') == string::npos
                           )
                        || (    event.key.keysym.unicode == '+'
                             && !input.size()
                           )
                        || (    event.key.keysym.unicode == '-'
                             && !input.size()
                           )
                      )                            
                    {
                        bmp_printf(   &BMP, x_off, y_off, BMP.make_rgb(0,255,255), bmp_bg_color, 0, font_size_factor
                                    , "%c"
                                    , event.key.keysym.unicode
                                  );
                        bmp_to_SDL_Surface(&bmp, screen);
                        input += event.key.keysym.unicode;
                        x_off += x_gap;
                    }                            
                    break;                         
                //-------------------------------------------------    
            } // end switch(event.key.keysym.unicode)
        } // end else of if(event.key.keysym.unicode == SDLK_RETURN)
    } // end while(!done)    
    bmp_printf(   &BMP
                , x_off - x_gap * input.size()
                , y_off
                , BMP.make_rgb(255,0,0)
                , bmp_bg_color
                , 0
                , font_size_factor
                , input.c_str()
              );
    bmp_to_SDL_Surface(&bmp, screen);
    sscanf(input.c_str(), "%f", &result);
    return result;
}

//############################################################################
int ez_uail_gui::display_prompt_and_echo_int(string prompt)
{
    bool   done         = false;    
    int    x_off        = EZ_STATS_DISPLAY_OFFSET * font_size_factor,
           y_off        = BMP.yres - 24 * font_size_factor,
           x_gap        = 8 * font_size_factor,
           y_line       = 8 * font_size_factor,
           result;
    u_int  bmp_bg_color = BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b);
    string input;
        
    process_exception = false;
    bmp_printf(   &BMP, x_off, y_off, BMP.white, bmp_bg_color, 0, font_size_factor
                , prompt.c_str()
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, BMP.make_rgb(0,255,255), bmp_bg_color, 0, font_size_factor
                , "_"
              );
    bmp_to_SDL_Surface(&bmp, screen);

    while(!done)
    {
        do
        {
            SDL_WaitEvent(&event);
        } while(event.type != SDL_KEYDOWN);
        if(event.key.keysym.unicode == SDLK_ESCAPE)
        {
            process_exception = true;
            return 0;
        }
        if(event.key.keysym.unicode == SDLK_RETURN)
            done = true;
        else
        {
            switch(event.key.keysym.unicode)
            {
                //-------------------------------------------------    
                case SDLK_DELETE:
                case SDLK_BACKSPACE:
                    if(input.size())
                    {
                        input = input.substr(0, input.size() - 1);
                        x_off -= x_gap;
                        bmp_printf(   &BMP, x_off, y_off, BMP.make_rgb(0,255,255), bmp_bg_color, 0, font_size_factor
                                    , " "
                                  );
                        bmp_to_SDL_Surface(&bmp, screen);
                    }                         
                    break;
                //-------------------------------------------------    
                default:
                    if(    isdigit(event.key.keysym.unicode)
                        || (    event.key.keysym.unicode == '+'
                             && !input.size()
                           )
                        || (    event.key.keysym.unicode == '-'
                             && !input.size()
                           )
                      )                    
                    {
                        bmp_printf(   &BMP, x_off, y_off, BMP.make_rgb(0,255,255), bmp_bg_color, 0, font_size_factor
                                    , "%c"
                                    , event.key.keysym.unicode
                                  );
                        bmp_to_SDL_Surface(&bmp, screen);
                        input += event.key.keysym.unicode;
                        x_off += x_gap;
                    }                            
                    break;                         
                //-------------------------------------------------    
            } // end switch(event.key.keysym.unicode)
        } // end else of if(event.key.keysym.unicode == SDLK_RETURN)
    } // end while(!done)
    bmp_printf(   &BMP
                , x_off - x_gap * input.size()
                , y_off
                , BMP.make_rgb(255,0,0)
                , bmp_bg_color
                , 0
                , font_size_factor
                , input.c_str()
              );
    bmp_to_SDL_Surface(&bmp, screen);
    if(input.size())
        sscanf(input.c_str(), "%d", &result);
    else
        result = 0;
    return result;
}

//############################################################################
int ez_uail_gui::display_prompt_and_echo_nibble(string prompt)
{
    bool   done         = false;    
    int    x_off        = EZ_STATS_DISPLAY_OFFSET * font_size_factor,
           y_off        = BMP.yres - 24 * font_size_factor,
           y_line       = 8 * font_size_factor,
           result;
    u_int  bmp_bg_color = BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b);
        
    process_exception = false;
    
    bmp_printf(   &BMP, x_off, y_off, BMP.white, bmp_bg_color, 0, font_size_factor
                , prompt.c_str()
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, BMP.make_rgb(0,255,255), bmp_bg_color, 0, font_size_factor
                , "_"
              );
    bmp_to_SDL_Surface(&bmp, screen);

    while(!done)
    {
        do
        {
            SDL_WaitEvent(&event);
        } while(event.type != SDL_KEYDOWN);
        if(event.key.keysym.unicode == SDLK_ESCAPE)
        {
            process_exception = true;
            return -1;
        }
        switch(event.key.keysym.unicode)
        {
            case '0':    return 0x00;
            case '1':    return 0x01;
            case '2':    return 0x02;
            case '3':    return 0x03;
            case '4':    return 0x04;
            case '5':    return 0x05;
            case '6':    return 0x06;
            case '7':    return 0x07;
            case '8':    return 0x08;
            case '9':    return 0x09;
            case 'a':    return 0x0a;
            case 'b':    return 0x0b;
            case 'c':    return 0x0c;
            case 'd':    return 0x0d;
            case 'e':    return 0x0e;
            case 'f':    return 0x0f;
            //-------------------------------------------------    
        } // end switch(event.key.keysym.unicode)
    } // end while(!done)
    return result;
}

//############################################################################
bool ez_uail_gui::display_prompt_and_echo_bool(string prompt)
{
    int    x_off  = EZ_STATS_DISPLAY_OFFSET * font_size_factor,
           y_off  = BMP.yres - 24 * font_size_factor;

    process_exception = false;
    bmp_printf(   &BMP, x_off, y_off, BMP.white, BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b), 0, font_size_factor
                , prompt.c_str()
              );
    bmp_to_SDL_Surface(&bmp, screen);    
    do
    {
        SDL_WaitEvent(&event);
    } while(event.type != SDL_KEYDOWN);
    if(event.key.keysym.unicode == SDLK_ESCAPE)
    {
        process_exception = true;
        return false;
    }
    if(event.key.keysym.unicode == 'y')
        return true;
    return false;
}

//############################################################################
void ez_uail_gui::display_program_menu()
{
    int    y_line       = 8 * font_size_factor,
           x_off        = EZ_STATS_DISPLAY_OFFSET * font_size_factor,
           y_off        = x_off;
    u_int  bmp_bg_color = BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b);

    bmp_printf(   &BMP, x_off, y_off, BMP.white, bmp_bg_color, 0, font_size_factor
                , "Welcome to " LASERBOY_VERSION " !!!"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      i open file in"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      o save file out"       
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      t signal bit resolution"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      u user interface visuals"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      v view keys & apply"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    ` ~ auto show"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    - _ auto orbit"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    = + auto show & orbit"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    %c %c frame"
                , 0x1b
                , 0x1a
              );

    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    %c %c target palette"
                , 0x18
                , 0x19
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      9 first frame"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      0 last frame"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      ( set | unset frames-( select"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      ) set | unset )-frames select"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "[Space] select | unselect frame"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "[Bkspc] delete current frame"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      p palette   transforms"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      h frame     transforms"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      j frame set transforms"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      k draw, color, blank  (vertices)"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      l move, scale, rotate (segments)"
              );
    if(bg_file_name != ".")
    {
        y_off += y_line;
        bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                    , "      ; show | hide background bitmap"
                  );
    }
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      b LaserBoy wave utilities"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    / ? frame set stats | frame stats"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "  [Tab] system settings"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "  [Esc] quit LaserBoy, saving OUT.ild"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "[Enter] show | hide help"
              );
    bmp_to_SDL_Surface(&bmp, screen);
    return;
}

//############################################################################
void ez_uail_gui::display_please_wait()
{
    int x_off  = EZ_STATS_DISPLAY_OFFSET * font_size_factor,
        y_off  = x_off,
        y_line = 8 * font_size_factor;

    bmp_printf(   &BMP, x_off, y_off, BMP.white, BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b), 0, font_size_factor
                , "please wait..."
              );

    bmp_to_SDL_Surface(&bmp, screen);
    return;
}

//############################################################################
void ez_uail_gui::display_in_file_type_menu()
{
    int    y_line       = 8 * font_size_factor,
           x_off        = EZ_STATS_DISPLAY_OFFSET * font_size_factor,
           y_off        = x_off;
    u_int  bmp_bg_color = BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b);

    bmp_printf(   &BMP, x_off, y_off, BMP.white, bmp_bg_color, 0, font_size_factor
                , "file input type"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    1 ILDA"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    2 DXF"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    3 LaserBoy wave"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    4 ADAT wave set"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    5 bitmap"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    6 ascii palette"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, BMP.white, bmp_bg_color, 0, font_size_factor
                , "[Esc] cancel action"
              );

    bmp_to_SDL_Surface(&bmp, screen);
    return;
}

//############################################################################
void ez_uail_gui::display_in_ild_action_menu()
{
    int    y_line       = 8 * font_size_factor,
           x_off        = EZ_STATS_DISPLAY_OFFSET * font_size_factor,
           y_off        = x_off;
    u_int  bmp_bg_color = BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b);

    bmp_printf(   &BMP, x_off, y_off, BMP.white, bmp_bg_color, 0, font_size_factor
                , "ILDA file input action"
              );

    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    1 replace frame set"
              );

    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    2 prepend to frame set"
              );

    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    3 insert after current frame"
              );

    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    4 append to frame set"
              );

    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    5 superimpose frame(s) into frame set"
              );

    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, BMP.white, bmp_bg_color, 0, font_size_factor
                , "[Esc] cancel action"
              );

    bmp_to_SDL_Surface(&bmp, screen);
    return;
}

//############################################################################
void ez_uail_gui::display_in_dxf_action_menu()
{
    int    y_line       = 8 * font_size_factor,
           x_off        = EZ_STATS_DISPLAY_OFFSET * font_size_factor,
           y_off        = x_off;
    u_int  bmp_bg_color = BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b);

    bmp_printf(   &BMP, x_off, y_off, BMP.white, bmp_bg_color, 0, font_size_factor
                , "DXF file input action"
              );

    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    1 replace frame set with DXF file"
              );

    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    2 append  frame set with DXF file"
              );

    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    3 replace frame set with DXF directory"
              );

    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    4 append  frame set with DXF directory"
              );

    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, BMP.white, bmp_bg_color, 0, font_size_factor
                , "[Esc] cancel action"
              );

    bmp_to_SDL_Surface(&bmp, screen);
    return;
}

//############################################################################
void ez_uail_gui::display_in_wave_action_menu()
{
    int    y_line       = 8 * font_size_factor,
           x_off        = EZ_STATS_DISPLAY_OFFSET * font_size_factor,
           y_off        = x_off;
    u_int  bmp_bg_color = BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b);

    bmp_printf(   &BMP, x_off, y_off, BMP.white, bmp_bg_color, 0, font_size_factor
                , "wave file input action"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    1 replace current frame set"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    2 append to current frame set"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, BMP.white, bmp_bg_color, 0, font_size_factor
                , "[Esc] cancel action"
              );

    bmp_to_SDL_Surface(&bmp, screen);
    return;
}

//############################################################################
void ez_uail_gui::display_in_bmp_action_menu()
{
    int    y_line       = 8 * font_size_factor,
           x_off        = EZ_STATS_DISPLAY_OFFSET * font_size_factor,
           y_off        = x_off;
    u_int  bmp_bg_color = BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b);

    bmp_printf(   &BMP, x_off, y_off, BMP.white, bmp_bg_color, 0, font_size_factor
                , "bitmap file input action"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    1 open as frame color map"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    2 open as frame set color map"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    3 open as frame mask"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    4 open as frame set mask"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    5 open into frame background"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    6 clear frame background"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, BMP.white, bmp_bg_color, 0, font_size_factor
                , "[Esc] cancel action"
              );

    bmp_to_SDL_Surface(&bmp, screen);
    return;
}

//############################################################################
void ez_uail_gui::display_if_not_z_options()
{
    int    y_line       = 8 * font_size_factor,
           x_off        = EZ_STATS_DISPLAY_OFFSET * font_size_factor,
           y_off        = x_off;
    u_int  bmp_bg_color = BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b);

    bmp_printf(   &BMP, x_off, y_off, BMP.white, bmp_bg_color, 0, font_size_factor
                , "signal option if not z"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    1 nothing"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    2 monochrome on or off"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    3 monochrome approximation of intensity"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, BMP.white, bmp_bg_color, 0, font_size_factor
                , "[Esc] cancel action"
              );

    bmp_to_SDL_Surface(&bmp, screen);
    return;
}

//############################################################################
void ez_uail_gui::display_out_file_type_menu()
{
    int    y_line       = 8 * font_size_factor,
           x_off        = EZ_STATS_DISPLAY_OFFSET * font_size_factor,
           y_off        = x_off;
    u_int  bmp_bg_color = BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b);

    bmp_printf(   &BMP, x_off, y_off, BMP.white, bmp_bg_color, 0, font_size_factor
                , "Output file type"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    1 ILDA save current frame"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    2 ILDA save selected frames"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    3 ILDA save all frames"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    4 ILDA generate & save frame effect"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    5 ILDA generate & save frame set effect"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    6 DXF save current frame"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    7 DXF save directory selected frames"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    8 DXF save directory all frames"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    9 wave animation %3.3f fps of all frames"
                , frames_per_second
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    0 wave %6.2f second still of current frame"
                , still_frame_duration
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    a C++ save current frame"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    b C++ save selected frames"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    c C++ save all frames"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    d bmp save %d%c pxl of current frame"
                , output_bmp_size
                , 0xfd
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    e bmp save directory %d%c pxl selected frames"
                , output_bmp_size
                , 0xfd
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    f bmp save directory %d%c pxl all frames"
                , output_bmp_size
                , 0xfd
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    g ASCII palette save current frame"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    h ASCII palette save target palette"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    i save system settings [LaserBoy.wtf]"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, BMP.white, bmp_bg_color, 0, font_size_factor
                , "[ESC] cancel action"
              );
    bmp_to_SDL_Surface(&bmp, screen);    
    return;
}

//############################################################################
void ez_uail_gui::display_settings_menu()
{
    int    y_line       = 8 * font_size_factor,
           x_off        = EZ_STATS_DISPLAY_OFFSET * font_size_factor,
           y_off        = x_off;
    u_int  bmp_bg_color = BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b);

    bmp_printf(   &BMP, x_off, y_off, BMP.white, bmp_bg_color, 0, font_size_factor
                , "LaserBoy settings"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    1 insignificant angle in degrees ..... %6.2f%c"
                , zero_angle_tolerance * one_radian
                , 0xf8                
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    2 insignificant distance in points ... %6.2f"
                , insignificant_distance
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    3 max optimized lit vector in points . %6.2f"
                , lit_distance_delta_max
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    4 max optimized blank vector in points %6.2f"
                , blank_distance_delta_max
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    5 maximum dwell in microseconds ...... %6.2f"
                , longest_dwell_microsec
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    6 lit dwell overhang in samples ...... %d"
                , lit_dwell_overhang
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    7 dwell on dot in samples ............ %d"
                , dwell_on_dot
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    8 move points per key hit ............ %d"
                , move_points_per_key
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    9 rotate degrees per key hit ......... %3.2f%c"
                , rotate_radians_per_key * one_radian
                , 0xf8
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    0 scale percent per key hit .......... %3.2f"
                , scale_per_key * 100.00
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    a rendered arc step in degrees ....... %3.2f%c"
                , rendered_arc_angle
                , 0xf8
              );              
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    b frames per second, 0 for once each . %6.2f"
                , frames_per_second
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    c still frame durration in seconds ... %6.2f"
                , still_frame_duration
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    d output wave sample rate ............ %d"
                , sample_rate
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    e wave signal if not z ............... %s"
                , ez_wave_signal_id_to_name((short) signal_if_not_z).c_str()
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    f output bitmap size in square pixels  %d"
                , output_bmp_size
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    g display font size factor ........... %d"
                , font_size_factor
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    h display background color ....... red %d"
                , bg_color.r
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    i display background color ..... green %d"
                , bg_color.g
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    j display background color ...... blue %d"
                , bg_color.b
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    k display background gray scale ...... %d"
                , (bg_color.r + bg_color.g + bg_color.b) / 3
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    l recolor palette span factor ........ %6.2f"
                , recolor_span_factor
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    m black level for palette conversion.. %d"
                , black_threshold
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "[ESC] exit settings menu"
              );
    bmp_to_SDL_Surface(&bmp, screen);    
    return;
}

//############################################################################
void ez_uail_gui::display_veiw_menu()
{
    int    x_off        = EZ_STATS_DISPLAY_OFFSET * font_size_factor,
           y_off        = x_off,
           y_line       =  8 * font_size_factor,
           x_tab        = 17 * y_line,
           x_tab2       =  9 * y_line;
    u_int  bmp_bg_color = BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b);

    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , bmp_bg_color
                , 0
                , font_size_factor
                , "press & hold view keys, [Shift] reverse"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.make_rgb(255,255,0)
                , bmp_bg_color
                , 0
                , font_size_factor
                , "1 ! orbit X"
              );
    bmp_printf(   &BMP
                , x_tab
                , y_off
                , BMP.make_rgb(0,255,255)
                , bmp_bg_color
                , 0
                , font_size_factor
                , "5 %% front | back"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.make_rgb(255,255,0)
                , bmp_bg_color
                , 0
                , font_size_factor
                , "2 @ orbit Y"
              );
    bmp_printf(   &BMP
                , x_tab
                , y_off
                , BMP.make_rgb(0,255,255)
                , bmp_bg_color
                , 0
                , font_size_factor
                , "6 ^  left | right"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.make_rgb(255,255,0)
                , bmp_bg_color
                , 0
                , font_size_factor
                , "3 # orbit Z"
              );
    bmp_printf(   &BMP
                , x_tab
                , y_off
                , BMP.make_rgb(0,255,255)
                , bmp_bg_color
                , 0
                , font_size_factor
                , "7 &   top | bottom"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.make_rgb(255,255,0)
                , bmp_bg_color
                , 0
                , font_size_factor
                , "4 $ orbit XYZ"
              );
    bmp_printf(   &BMP
                , x_tab
                , y_off
                , BMP.make_rgb(0,255,255)
                , bmp_bg_color
                , 0
                , font_size_factor
                , "8 * orbit & frame"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.make_rgb(255,0,255)
                , bmp_bg_color
                , 0
                , font_size_factor
                , "q Q pan X"
              );
    bmp_printf(   &BMP
                , x_tab
                , y_off
                , BMP.make_rgb(0,255,0)
                , bmp_bg_color
                , 0
                , font_size_factor
                , "a A zoom X"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.make_rgb(255,0,255)
                , bmp_bg_color
                , 0
                , font_size_factor
                , "w W pan Y"
              );
    bmp_printf(   &BMP
                , x_tab
                , y_off
                , BMP.make_rgb(0,255,0)
                , bmp_bg_color
                , 0
                , font_size_factor
                , "s S zoom Y"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.make_rgb(255,0,255)
                , bmp_bg_color
                , 0
                , font_size_factor
                , "e E pan XY"
              );
    bmp_printf(   &BMP
                , x_tab
                , y_off
                , BMP.make_rgb(0,255,0)
                , bmp_bg_color
                , 0
                , font_size_factor
                , "d D zoom XY"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.make_rgb(255,0,255)
                , bmp_bg_color
                , 0
                , font_size_factor
                , "r R center"
              );
    bmp_printf(   &BMP
                , x_tab
                , y_off
                , BMP.make_rgb(0,255,0)
                , bmp_bg_color
                , 0
                , font_size_factor
                , "f F 1:1"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , bmp_bg_color
                , 0
                , font_size_factor
                , "current view"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_tab2
                , y_off
                , BMP.make_rgb(255,0,0)
                , bmp_bg_color
                , 0
                , font_size_factor
                , "X"
              );
    bmp_printf(   &BMP
                , x_tab2 * 2
                , y_off
                , BMP.make_rgb(0,255,0)
                , bmp_bg_color
                , 0
                , font_size_factor
                , "Y"
              );
    bmp_printf(   &BMP
                , x_tab2 * 3
                , y_off
                , BMP.make_rgb(0,0,255)
                , bmp_bg_color
                , 0
                , font_size_factor
                , "Z"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.make_rgb(255,255,0)
                , bmp_bg_color
                , 0
                , font_size_factor
                , " angle                               degrees"
              );
    bmp_printf(   &BMP
                , x_tab2
                , y_off
                , BMP.make_rgb(255,255,0)
                , BMP.make_rgb(80,0,0)
                , 0
                , font_size_factor
                , "%-3.4f"
                , view_angle.x * one_radian
              );
    bmp_printf(   &BMP
                , x_tab2 * 2
                , y_off
                , BMP.make_rgb(255,255,0)
                , BMP.make_rgb(0,80,0)
                , 0
                , font_size_factor
                , "%-3.4f"
                , view_angle.y * one_radian
              );
    bmp_printf(   &BMP
                , x_tab2 * 3
                , y_off
                , BMP.make_rgb(255,255,0)
                , BMP.make_rgb(0,0,80)
                , 0
                , font_size_factor
                , "%-3.4f"
                , view_angle.z * one_radian
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.make_rgb(255,0,255)
                , bmp_bg_color
                , 0
                , font_size_factor
                , "offset                               points"
              );
    bmp_printf(   &BMP
                , x_tab2
                , y_off
                , BMP.make_rgb(255,0,255)
                , BMP.make_rgb(80,0,0)
                , 0
                , font_size_factor
                , "%-8.0f"
                , view_offset.x
              );
    bmp_printf(   &BMP
                , x_tab2 * 2
                , y_off
                , BMP.make_rgb(255,0,255)
                , BMP.make_rgb(0,80,0)
                , 0
                , font_size_factor
                , "%-8.0f"
                , view_offset.y
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.make_rgb(0,255,0)
                , bmp_bg_color
                , 0
                , font_size_factor
                , " scale"
              );
    bmp_printf(   &BMP
                , x_tab2
                , y_off
                , BMP.make_rgb(0,255,0)
                , BMP.make_rgb(80,0,0)
                , 0
                , font_size_factor
                , "%-3.4f"
                , view_scale.x
              );
    bmp_printf(   &BMP
                , x_tab2 * 2
                , y_off
                , BMP.make_rgb(0,255,0)
                , BMP.make_rgb(0,80,0)
                , 0
                , font_size_factor
                , "%-3.4f"
                , view_scale.y
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.make_rgb(255,255,0)
                , bmp_bg_color
                , 0
                , font_size_factor
                , "      o O apply angle  to frame | all"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.make_rgb(255,0,255)
                , bmp_bg_color
                , 0
                , font_size_factor
                , "      t T apply offset to frame | all"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.make_rgb(0,255,0)
                , bmp_bg_color
                , 0
                , font_size_factor
                , "      g G apply scale  to frame | all"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , bmp_bg_color
                , 0
                , font_size_factor
                , "  [Esc] exit view settings"
              );
    bmp_to_SDL_Surface(&bmp, screen);
    return;
}

//############################################################################
void ez_uail_gui::display_visuals_menu()
{
    int    y_line       = 8 * font_size_factor,
           x_off        = EZ_STATS_DISPLAY_OFFSET * font_size_factor,
           y_off        = x_off;
    u_int  bmp_bg_color = BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b);

    bmp_printf(   &BMP, x_off, y_off, BMP.white, bmp_bg_color, 0, font_size_factor
                , "LaserBoy toggle UI visuals settings"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , ((show_points)?("--> 1 vertices"):("    1 vertices"))
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , ((show_blanking)?("--> 2 blanking"):("    2 blanking"))
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , ((show_palettes)?("--> 3 palettes"):("    3 palettes"))
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , ((show_stats)?("--> 4 stats"):("    4 stats"))
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , ((show_fixed_bounds)?("--> 5 fixed bounds"):("    5 fixed bounds"))
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , ((show_fixed_origin)?("--> 6 fixed origin"):("    6 fixed origin"))
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , ((show_floating_axis)?("--> 7 floating axis"):("    7 floating axis"))
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , ((show_floating_bounds)?("--> 8 floating bounds"):("    8 floating bounds"))
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , ((show_ui_message)?("--> 9 UI message"):("    9 UI message"))
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    0 show | hide all visuals"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, BMP.white, bmp_bg_color, 0, font_size_factor
                , "toggle LaserBoy system settings"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , ((show_effects_generation)?("--> a show effect rendering"):("    a show effect rendering"))
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , ((auto_flatten_z)?("--> b flatten Z for wave output"):("    b flatten Z for wave output"))
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , ((no_equivalent_vectors)?("--> c remove equal vectors"):("    c remove equal vectors"))
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , ((no_dots)?("--> d remove dots (not lines)"):("    d remove dots (not lines)"))
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , ((invert_wave_output)?("--> e inverted wave output"):("    e inverted wave output"))
              );              
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , ((auto_optimize)?("--> f optimize before save"):("    f optimize before save"))
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , ((loop_animation)?("--> g loop animation"):("    g loop animation"))
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , ((approximate_frame_rate)?("--> h animate at approximated frame rate"):("    h animate at approximated frame rate"))
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , ((auto_scale_dxf)?("--> i auto scale DXF import"):("    i auto scale DXF import"))
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , ((maintain_dxf_origin)?("--> j maintain DXF origin"):("    j maintain DXF origin"))
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , ((black_dwell_points)?("--> k black dwell"):("    k black dwell"))
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , ((no_black_match)?("--> l no black palette match"):("    l no black palette match"))
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , ((allow_lit_black)?("--> m allow lit black"):("    m allow lit black"))
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , ((dumb_file_format)?("--> n save ilda format 4,5"):("    n save ilda format 4,5"))
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, BMP.white, bmp_bg_color, 0, font_size_factor
                , "[Esc] exit LaserBoy UI visuals settings"
              );
    bmp_to_SDL_Surface(&bmp, screen);
    return;
}

//############################################################################
void ez_uail_gui::display_palette_tranz_menu()
{
    int    y_line       = 8 * font_size_factor,
           x_off        = EZ_STATS_DISPLAY_OFFSET * font_size_factor,
           y_off        = x_off;
    u_int  bmp_bg_color = BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b);

    bmp_printf(   &BMP, x_off, y_off, BMP.white, bmp_bg_color, 0, font_size_factor
                , "palette transforms"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      1 span X with palette"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      2 span Y with palette"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      3 span Z with palette"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      4 span  vectors with palette"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      5 index vectors with palette"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      6 span  lit segments with palette"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      7 index lit segments with palette"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      8 random color vectors with palette"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      9 random color lit segments with palette"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      0 set to target palette by index"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      a best match to target palette"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      b convert true color to palette (if possible)"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      c bit reduce true color to 332 palette"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      d best reduce true color to palette"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      e promote to true color"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      f F best reduce all frames to 1 palette"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      g strip color information"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "[Shift] select to apply to all frames"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, BMP.white, bmp_bg_color, 0, font_size_factor
                , "  [Esc] exit palette transforms"
              );
    bmp_to_SDL_Surface(&bmp, screen);
    return;
}

//############################################################################
void ez_uail_gui::display_frame_set_tranz_menu()
{
    int    y_line       = 8 * font_size_factor,
           x_off        = EZ_STATS_DISPLAY_OFFSET * font_size_factor,
           y_off        = x_off;
    u_int  bmp_bg_color = BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b);

    bmp_printf(   &BMP, x_off, y_off, BMP.white, bmp_bg_color, 0, font_size_factor
                , "frame set transforms"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      1 delete selected"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      2 trim to selected"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      3 delete repeats from wave import"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      4 copy selected to begining"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      5 copy selected to end"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      6 move selected to begining"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      7 move selected to end"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      8 add empty frame to begining"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      9 add empty frame after current"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      0 add empty frame to end"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      a reverse frame set order"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      b reverse order of selected frames"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      c make current frame first"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      d randomize frame set order"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      e delete every n-th frame"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      > explode frame into segment set"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      < collapse selected frames into one"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      ! explode all frames into segments"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "[Space] select | unselect frame"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "[Enter] show | hide help"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, BMP.white, bmp_bg_color, 0, font_size_factor
                , "  [Esc] exit frame set transforms"
              );
    bmp_to_SDL_Surface(&bmp, screen);
    return;
}

//############################################################################
void ez_uail_gui::display_frame_tranz_menu()
{
    int    y_line       = 8 * font_size_factor,
           x_off        = EZ_STATS_DISPLAY_OFFSET * font_size_factor,
           y_off        = x_off;
    u_int  bmp_bg_color = BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b);

    bmp_printf(   &BMP, x_off, y_off, BMP.white, bmp_bg_color, 0, font_size_factor
                , "frame transforms"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      1 flatten Z (force to 2D)"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      2 Z order vertices"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      3 reverse vector order"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      4 mirror"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      5 flip"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      6 quarter turn CW"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      7 quarter turn CCW"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      8 accentuate the positive"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      9 eliminate the negative"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      0 convert black to blank"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      a convert to dots"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      b remove blank span"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      c remove angle dwell"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      d remove dots (not lines)"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      e reduce vectors"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      f blank equivalent vectors"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      g bond segments (reorder)"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      h minimize"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      i fracture segments"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      j reorder segments"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      k add angle dwell"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      l add lit distance span"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      m add blank distance span"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      n enhance dots"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      o optimize"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "[Shift] select applies to all frames"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "[Enter] show | hide help"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, BMP.white, bmp_bg_color, 0, font_size_factor
                , "  [Esc] exit frame transforms"
              );
    bmp_to_SDL_Surface(&bmp, screen);
    return;
}

//############################################################################
void ez_uail_gui::display_draw_color_blank_menu()
{
    int    y_line       = 8 * font_size_factor,
           x_off        = EZ_STATS_DISPLAY_OFFSET * font_size_factor,
           y_off        = x_off;
    u_int  bmp_bg_color = BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b);

    bmp_printf(   &BMP, x_off, y_off, BMP.white, bmp_bg_color, 0, font_size_factor
                , "draw, color, blank mode entered"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    \\ | %c to 0 | %c to end"
                , 0x01
                , 0x0f
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    [ { +- %c   one vertex"
                , 0x01
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    ] } +-   %c one vertex"
                , 0x0f
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    i I +- %c   lit segment"
                , 0x01
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    o O +-   %c lit segment"
                , 0x0f
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    u U +- %c-%c lit segment"
                , 0x01
                , 0x0f
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    a A unblank all vectors | frames"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    b B blank | unblank %c-%c"
                , 0x01
                , 0x0f
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    n N bond black | unbond unblack %c-%c"
                , 0x01
                , 0x0f
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    p P +- palette color index"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      c color %c-%c with palette index"
                , 0x01
                , 0x0f
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      C get palette index from %c color"
                , 0x0f
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      h replace black in %c-%c with palette index"
                , 0x01
                , 0x0f
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    X Y Z  move   %c + %d"
                , 0x0f
                , move_points_per_key
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    X Y Z  move   %c - %d"
                , 0x0f
                , move_points_per_key
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    q w e  set    %c coordinates X Y Z"
                , 0x0f
                , move_points_per_key
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    f F    rotate %c +- %3.2f%c on Z axis"
                , 0x0f
                , rotate_radians_per_key * one_radian
                , 0xf8
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    g G    rotate %c +- %3.2f%c on Y axis"
                , 0x0f
                , rotate_radians_per_key * one_radian
                , 0xf8
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    m M +- %3.2f%% magnitude of %c"
                , scale_per_key * 100.00
                , 0x0f
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      l draw line    %c to %c"
                , 0x01
                , 0x0f
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      k draw polygon %c to %c radius"
                , 0x01
                , 0x0f
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      d %3.2f%c circular  arc, %c to %c radius"
                , rendered_arc_angle
                , 0xf8
                , 0x01
                , 0x0f
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      D %3.2f%c eliptical arc, %c to %c radius"
                , rendered_arc_angle
                , 0xf8
                , 0x01
                , 0x0f
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      t render text on line  %c to %c"
                , 0x01
                , 0x0f
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      ' convert %c-%c to dots"
                , 0x01
                , 0x0f
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      . split vector in line or add vector to end"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      s add blank vertex at %c (cut segment)"
                , 0x0f
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "[Bkspc] delete vertex at %c"
                , 0x0f
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    / ? show frame | set stats"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "%c %c %c %c and view orbit keys are active"
                , 0x1b
                , 0x1a
                , 0x18
                , 0x19
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "[Enter] show | hide help"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, BMP.white, bmp_bg_color, 0, font_size_factor
                , "  [Esc] exit draw, color blank mode"
              );
    bmp_to_SDL_Surface(&bmp, screen);
    return;
}

//############################################################################
void ez_uail_gui::display_move_scale_rotate_menu()
{
    int    y_line       = 8 * font_size_factor,
           x_off        = EZ_STATS_DISPLAY_OFFSET * font_size_factor,
           y_off        = x_off;
    u_int  bmp_bg_color = BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b);
    
    bmp_printf(   &BMP, x_off, y_off, BMP.white, bmp_bg_color, 0, font_size_factor
                , "move, scale, rotate mode entered"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    \\ | %c to 0 | %c to end"
                , 0x01
                , 0x0f
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    [ { +- %c   one vertex"
                , 0x01
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    ] } +-   %c one vertex"
                , 0x0f
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    i I +- %c   lit segment"
                , 0x01
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    o O +-   %c lit segment"
                , 0x0f
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    u U +- %c-%c lit segment"
                , 0x01
                , 0x0f
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    X Y Z  move   %c + %d"
                , 0x0f
                , move_points_per_key
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    X Y Z  move   %c - %d"
                , 0x0f
                , move_points_per_key
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    q w e  set center coordinates X Y Z"
                , 0x0f
                , move_points_per_key
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    f F rotate %c-%c +- %3.2f%c Z axis on center"
                , 0x01
                , 0x0f
                , rotate_radians_per_key * one_radian
                , 0xf8
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    g G rotate %c-%c +- %3.2f%c Y axis on center"
                , 0x01
                , 0x0f
                , rotate_radians_per_key * one_radian
                , 0xf8
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    9 ( scale  %c-%c X   +- %3.2f%% on center"
                , 0x01
                , 0x0f
                , scale_per_key * 100.00
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    0 ) scale  %c-%c  Y  +- %3.2f%% on center"
                , 0x01
                , 0x0f
                , scale_per_key * 100.00
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    - _ scale  %c-%c   Z +- %3.2f%% on center"
                , 0x01
                , 0x0f
                , scale_per_key * 100.00
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    = + scale  %c-%c XYZ +- %3.2f%% on center"
                , 0x01
                , 0x0f
                , scale_per_key * 100.00
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    a s d  move  fulcrum X Y Z + %d"
                , move_points_per_key
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    A S D  move  fulcrum X Y Z - %d"
                , move_points_per_key
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    j k l  set   fulcrum coordinates X Y Z"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    r R rotate %c-%c +- %3.2f%c Z axis on fulcrum"
                , 0x01
                , 0x0f
                , rotate_radians_per_key * one_radian
                , 0xf8
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    t T rotate %c-%c +- %3.2f%c Y axis on fulcrum"
                , 0x01
                , 0x0f
                , rotate_radians_per_key * one_radian
                , 0xf8
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    v V scale  %c-%c X   +- %3.2f%% from fulcrum"
                , 0x01
                , 0x0f
                , scale_per_key * 100.00
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    b B scale  %c-%c  Y  +- %3.2f%% from fulcrum"
                , 0x01
                , 0x0f
                , scale_per_key * 100.00
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    n N scale  %c-%c   Z +- %3.2f%% from fulcrum"
                , 0x01
                , 0x0f
                , scale_per_key * 100.00
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    m M scale  %c-%c XYZ +- %3.2f%% from fulcrum"
                , 0x01
                , 0x0f
                , scale_per_key * 100.00
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      c copy   %c-%c (in place) to end"
                , 0x01
                , 0x0f
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    / ? frame | set stats"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "%c %c %c %c and view orbit keys are active"
                , 0x1b
                , 0x1a
                , 0x18
                , 0x19
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "[Enter] show | hide help"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, BMP.white, bmp_bg_color, 0, font_size_factor
                , "  [ESC] exit move, scale, rotate mode"
              );
    bmp_to_SDL_Surface(&bmp, screen);
    return;
}

//############################################################################
void ez_uail_gui::display_wave_utilities_menu()
{
    int y_line = 8 * font_size_factor,
        x_off  = EZ_STATS_DISPLAY_OFFSET * font_size_factor,
        y_off  = x_off;
    u_int  bmp_bg_color = BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b);
    
    bmp_printf(   &BMP, x_off, y_off, BMP.white, bmp_bg_color, 0, font_size_factor
                , "LaserBoy wave utilities"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    1 LaserBoy wave (screen) show settings"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    2 show LaserBoy wave header information"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    3 show LaserBoy wave on screen"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    4 set desired LaserBoy wave sample offsets"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    5 apply offsets to LaserBoy wave"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    6 overwrite offsets in LaserBoy wave header"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    7 invert polarity of LaserBoy wave"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    8 split LaserBoy wave to XY, R, G, B, I"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    9 join XY, R, G, B, I waves to LaserBoy wave"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    0 add stereo audio to LaserBoy wave"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "[ESC] exit LaserBoy wave utilities"
              );
    bmp_to_SDL_Surface(&bmp, screen);
    return;
}

//############################################################################
void ez_uail_gui::display_wave_show_settings_menu()
{
    int    y_line       = 8 * font_size_factor,
           x_off        = EZ_STATS_DISPLAY_OFFSET * font_size_factor,
           y_off        = x_off;
    u_int  bmp_bg_color = BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b);

    bmp_printf(   &BMP, x_off, y_off, BMP.white, bmp_bg_color, 0, font_size_factor
                , "LaserBoy wave show settings"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , ((wave_show_points)?("--> 1 vertices"):("    1 vertices"))
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , ((wave_show_blanks)?("--> 2 blanking"):("    2 blanking"))
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , ((wave_real_time)?("--> 3 real time"):("    3 real time"))
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , ((wave_show_inverted)?("--> 4 inverted"):("    4 inverted"))
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , ((wave_clean_screen)?("--> 5 clean screen"):("    5 clean screen"))
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , ((wave_loop)?("--> 6 loop"):("    6 loop"))
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, BMP.white, bmp_bg_color, 0, font_size_factor
                , "[Esc] exit LaserBoy wave show settings"
              );
    bmp_to_SDL_Surface(&bmp, screen);
    return;
}

//############################################################################
void ez_uail_gui::display_wave_stats(string file_name)
{
    int            i,
                   y_line       = 8 * font_size_factor,
                   x_off        = EZ_STATS_DISPLAY_OFFSET * font_size_factor,
                   y_off        = x_off;
    u_int          bmp_bg_color = BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b);
    ifstream       in;
    //------------------------------------------------------------------------
    process_exception = false;
    //------------------------------------------------------------------------
    if(    file_name.size() < 4
        || file_name.substr(file_name.size() - 4) != ".wav"
      )
        file_name += ".wav";
    //------------------------------------------------------------------------
    in.open((EZ_UAIL_WAVE_SHARE + file_name).c_str(), ios::in | ios::binary);
    ez_wave_header header(in);
    in.close();
    if(header.version == "!LaserBoy!")
    {
        bmp_printf(   &BMP, x_off, y_off, BMP.white, bmp_bg_color, 0, font_size_factor
                    , "[any key] %s failed to open"
                    , file_name.c_str()
                  );
        error = EZ_UAIL_FILE_OPEN_FAILED;                  
        bmp_to_SDL_Surface(&bmp, screen);
        process_exception = true;
        return;
    }
    //------------------------------------------------------------------------
    bmp_printf(   &BMP, x_off, y_off, BMP.white, bmp_bg_color, 0, font_size_factor
                , "LaserBoy wave stats"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , (EZ_UAIL_WAVE_SHARE + file_name).c_str()
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "LaserBoy wave header version : %s"
                , header.version.c_str()
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "num_samples ..........: %d"
                , header.num_samples
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "sample_rate ..........: %d"
                , header.sample_rate
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "number of channels ...: %d"
                , header.num_channels
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "bits per sample ......: %d"
                , header.bits_per_sample
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "playing time .........: %f seconds"
                , ((double)header.num_samples / header.sample_rate)
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , ((header.ez_wave_mode & EZ_WAVE_POSITIVE) ? ("wave polarity ........: positive signal") : ("wave polarity ........: negative signal"))
              );              
    if(header.ez_wave_mode & EZ_WAVE_OPTIMIZED)
    {              
        y_off += y_line;
        bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                    , "lit dwell overhang ...: %d"
                    , header.parms.lit_dwell_overhang
                  );
        y_off += y_line;
        bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                    , "lit distance delta max: %6.2f"
                    , header.parms.lit_distance_delta_max
                  );
        y_off += y_line;
        bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                    , "blank dist delta max .: %6.2f"
                    , header.parms.blank_distance_delta_max
                  );
        y_off += y_line;
        bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                    , "longest dwell u_sec ..: %6.2f"
                    , header.parms.longest_dwell_microsec
                  );
        y_off += y_line;
        bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                    , "insignificant distance: %6.2f"
                    , header.parms.insignificant_distance
                  );
        y_off += y_line;
        bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                    , "insignificant angle ..: %6.2f"
                    , header.parms.zero_angle_tolerance
                  );
        y_off += y_line;
        bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                    , "frames per second ....: %6.2f"
                    , header.parms.frames_per_second
                  );
    }
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, BMP.white, bmp_bg_color, 0, font_size_factor
                , "ch: sig : contents     : LSB subcode  : offset : res"
              );
    for(i = 0; i < header.num_channels; i++)
    {
        y_off += y_line;
        bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                    , "%2d: %s : %s : %s :   %4d : %3d"
                    , i
                    , ((header.signal_id[i] >= 0) ? ("pos") : ("neg"))
                    , ez_wave_signal_id_to_name(header.signal_id[i]).c_str()
                    , ez_wave_LSB_tag_to_name(header.LSB_tag[i]).c_str()
                    , header.offset[i]
                    , header.resolution[i]
                  );
    }
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, BMP.white, bmp_bg_color, 0, font_size_factor
                , "press any key to continue"
                , ((double)header.num_samples / header.sample_rate)
              );
    bmp_to_SDL_Surface(&bmp, screen);
    return;
}

//############################################################################
void ez_uail_gui::display_wave(string file_name)
{
    bool                     playing      = true         ;
    int                      i                           ,
                             error                       ,
                             span                        ,
                             last                        ,
                             red                         ,
                             green                       ,
                             blue                        ,
                             sample_count = 0            ,
                             offset_max                  ,
                             offset_min                  ,
                             center       = screen->h / 2;
    u_int                    time_zero                   ,
                             color        = BMP.white    ,
                             bmp_bg_color = BMP.make_rgb(bg_color.r,
                                                         bg_color.g,
                                                         bg_color.b
                                                        );
    float                    scale  = screen->h / 65536.0;
    ifstream                 in                          ;
    struct  ezfb_bitmap      frame_buffer = {0}          ;
    ez_wave_sample_pointer   roll_over                   ;
    ez_wave_sample_pointer*  sample_window               ;
    //------------------------------------------------------------------------
    process_exception = false;
    //------------------------------------------------------------------------
    do // clear key strokes from event queue!
    {
        SDL_PollEvent(&event);    
    } while(event.type == SDL_KEYDOWN);
    //------------------------------------------------------------------------
    if(    file_name.size() < 4
        || file_name.substr(file_name.size() - 4) != ".wav"
      )
        file_name += ".wav";
    in.open((EZ_UAIL_WAVE_SHARE + file_name).c_str(), ios::in | ios::binary);
    //------------------------------------------------------------------------
    if(!in.is_open())
    {
        bmp_printf(   &BMP
                    , EZ_STATS_DISPLAY_OFFSET * font_size_factor
                    , EZ_STATS_DISPLAY_OFFSET * font_size_factor
                    , BMP.white
                    , bmp_bg_color
                    , 0
                    , font_size_factor
                    , "[any key] %s failed to open                "
                    , file_name.c_str()
                  );
        error = EZ_UAIL_FILE_OPEN_FAILED;
        bmp_to_SDL_Surface(&bmp, screen);
        process_exception = true;
        return;
    }
    //------------------------------------------------------------------------
    ez_wave_header header(in);
    if(header.version == "!LaserBoy!")
    {
        in.close();              
        bmp_printf(   &BMP
                    , EZ_STATS_DISPLAY_OFFSET * font_size_factor
                    , EZ_STATS_DISPLAY_OFFSET * font_size_factor
                    , BMP.white
                    , bmp_bg_color
                    , 0
                    , font_size_factor
                    , "[any key] %s wave header read failed       "
                  );
        bmp_to_SDL_Surface(&bmp, screen);
        process_exception = true;
        return;
    }
    //------------------------------------------------------------------------
    if(header.num_channels < 6)
    {
        in.close();              
        bmp_printf(   &BMP
                    , EZ_STATS_DISPLAY_OFFSET * font_size_factor
                    , EZ_STATS_DISPLAY_OFFSET * font_size_factor
                    , BMP.white
                    , bmp_bg_color
                    , 0
                    , font_size_factor
                    , "[any key] %s is not a LaserBoy wave        "
                    , file_name.c_str()
                  );
        error = EZ_UAIL_INCOMPATIBLE_FORMATS;
        bmp_to_SDL_Surface(&bmp, screen);
        process_exception = true;
        return;
    }
    //------------------------------------------------------------------------    
    offset_min = header.offset[0];
    offset_max = header.offset[0];    
    for(i = 1; i < 6; i++)
    {
        if(header.offset[i] > offset_max)
            offset_max = header.offset[i];
        if(header.offset[i] < offset_min)
            offset_min = header.offset[i];
    }
    //--------------------------------------------------------------------
    span  = (offset_max - offset_min) + 2; // at least 2 samples!
    last  = span - 1;
    red   = -header.offset[2] + 1;
    green = -header.offset[3] + 1;
    blue  = -header.offset[4] + 1;
    //------------------------------------------------------------------------
    sample_window = new ez_wave_sample_pointer[span]; // at least 2
    for(i = 0; i < span; i++)
        sample_window[i] = new ez_wave_sample(header.num_channels);
    //------------------------------------------------------------------------    
    bmp_clone_SDL_Surface(&frame_buffer, screen, 0);
    bmp_uint(&frame_buffer, bg_color.as_uint());
    //------------------------------------------------------------------------    
    while(playing)
    {
        time_zero = SDL_GetTicks();
        //----------------------------------------------------------------
        roll_over = sample_window[0];
        for(i = 0; i < last; i++)
            sample_window[i] = sample_window[i + 1];
        sample_window[last] = roll_over;
        //----------------------------------------------------------------
        if(wave_show_inverted)
            playing = sample_window[last]->from_ifstream_wave_inverted(in);
        else
            playing = sample_window[last]->from_ifstream_wave(in);
        //----------------------------------------------------------------
        sample_count++;        
        //----------------------------------------------------------------
        color = frame_buffer.make_rgb(   sample_window[red  ]->channel[2] >> 7
                                       , sample_window[green]->channel[3] >> 7
                                       , sample_window[blue ]->channel[4] >> 7
                                     );
        //----------------------------------------------------------------
        if(sample_window[red]->channel[2] & 0x0001)
            color = 0;
        //----------------------------------------------------------------
        if(color)
        {
            bmp_put_line_fast(   &frame_buffer
                               , (u_int)((sample_window[0]->channel[0] *  scale) + center)
                               , (u_int)((sample_window[0]->channel[1] * -scale) + center)
                               , (u_int)((sample_window[1]->channel[0] *  scale) + center)
                               , (u_int)((sample_window[1]->channel[1] * -scale) + center)
                               , color
                             );

            if(wave_show_points)
                bmp_put_rectangle(   &frame_buffer
                                   , (u_int)((sample_window[1]->channel[0] *  scale) + center - 1)
                                   , (u_int)((sample_window[1]->channel[1] * -scale) + center - 1)
                                   , (u_int)((sample_window[1]->channel[0] *  scale) + center + 1)
                                   , (u_int)((sample_window[1]->channel[1] * -scale) + center + 1)
                                   , color
                                   , false
                                 );
        }
        //----------------------------------------------------------------
        else if(wave_show_blanks)
        {
            bmp_put_line_fast(   &frame_buffer
                               , (u_int)((sample_window[0]->channel[0] *  scale) + center)
                               , (u_int)((sample_window[0]->channel[1] * -scale) + center)
                               , (u_int)((sample_window[1]->channel[0] *  scale) + center)
                               , (u_int)((sample_window[1]->channel[1] * -scale) + center)
                               , frame_buffer.gray
                             );

            if(wave_show_points)
                bmp_put_rectangle(   &frame_buffer
                                   , (u_int)((sample_window[1]->channel[0] *  scale) + center - 1)
                                   , (u_int)((sample_window[1]->channel[1] * -scale) + center - 1)
                                   , (u_int)((sample_window[1]->channel[0] *  scale) + center + 1)
                                   , (u_int)((sample_window[1]->channel[1] * -scale) + center + 1)
                                   , frame_buffer.gray
                                   , false
                                  );
        }
        //----------------------------------------------------------------
        if(sample_window[red]->channel[2] & 0x0001)
        {
            if(wave_real_time)
            {
                do
                {    
                     SDL_Delay(1);
                }
                while(   SDL_GetTicks() - time_zero
                       < (u_int)(sample_count / (header.sample_rate / 1000.0))
                     );                             
            }
            bmp_to_SDL_Surface(&frame_buffer, screen);
            if(wave_clean_screen)
                bmp_uint(&frame_buffer, bg_color.as_uint());
            SDL_PollEvent(&event);
            if(event.type == SDL_KEYDOWN)
                playing = false;
            sample_count = 0;
        }
        //----------------------------------------------------------------
    } // end while(playing)
    in.close();
    bmp_free(&frame_buffer);
    delete sample_window;
    //------------------------------------------------------------------------
    return;
}

//############################################################################
void ez_uail_gui::display_wave_offsets()
{
    int    y_line       = 8 * font_size_factor,
           x_off        = EZ_STATS_DISPLAY_OFFSET * font_size_factor,
           y_off        = x_off;
    u_int  bmp_bg_color = BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b);

    bmp_printf(   &BMP, x_off, y_off, BMP.white, bmp_bg_color, 0, font_size_factor
                , "LaserBoy wave sample offset settings"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "  [0] X           %d"
                , wave_offsets[0]
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "  [1] Y           %d"
                , wave_offsets[1]
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "  [2] red         %d"
                , wave_offsets[2]
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "  [3] green       %d"
                , wave_offsets[3]
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "  [4] blue        %d"
                , wave_offsets[4]
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "  [5] Z or other  %d"
                , wave_offsets[5]
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "  [6] audio left  %d"
                , wave_offsets[6]
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "  [7] audio right %d"
                , wave_offsets[7]
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, BMP.white, bmp_bg_color, 0, font_size_factor
                , "[Esc] cancel"
              );
    bmp_to_SDL_Surface(&bmp, screen);
    return;
}

//############################################################################
void ez_uail_gui::display_bit_resolution_settings()
{
    int    y_line       = 8 * font_size_factor,
           x_off        = EZ_STATS_DISPLAY_OFFSET * font_size_factor,
           y_off        = x_off;
    u_int  bmp_bg_color = BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b);

    bmp_printf(   &BMP, x_off, y_off, BMP.white, bmp_bg_color, 0, font_size_factor
                , "LaserBoy signal bit resolution settings"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "  [0] X           %d"
                , 16 - signal_bit_mask[0]
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "  [1] Y           %d"
                , 16 - signal_bit_mask[1]
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "  [5] Z or other  %d"
                , 16 - signal_bit_mask[5]
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "  [2] red         %d"
                , 16 - signal_bit_mask[2]
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "  [3] green       %d"
                , 16 - signal_bit_mask[3]
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "  [4] blue        %d"
                , 16 - signal_bit_mask[4]
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "  [6] audio left  %d"
                , 16 - signal_bit_mask[6]
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "  [7] audio right %d"
                , 16 - signal_bit_mask[7]
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, BMP.white, bmp_bg_color, 0, font_size_factor
                , "[Esc] cancel"
              );
    bmp_to_SDL_Surface(&bmp, screen);
    return;
}

//############################################################################
void ez_uail_gui::display_file_list(string directory)
{
    bool            found_one    = false;
    int             i,
                    y_line       = 8 * font_size_factor,
                    x_off        = BMP.yres + EZ_STATS_DISPLAY_OFFSET * font_size_factor,
                    y_off        = y_line;
    u_int           bmp_bg_color = BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b);
    DIR            *pDIR;
    struct dirent  *pDirEnt;

    pDIR = opendir(directory.c_str());
    if(pDIR == NULL)
        return;
    bmp_printf(   &BMP
                , x_off
                , 0
                , BMP.white
                , bmp_bg_color
                , 0
                , font_size_factor
                , "                  "
              );
    bmp_printf(   &BMP, x_off, y_off, BMP.white, bmp_bg_color, 0, font_size_factor
                , "                  "
              );
    bmp_printf(   &BMP, x_off, y_off, BMP.white, bmp_bg_color, 0, font_size_factor
                , "Palette set stats "
              );
    pDirEnt = readdir(pDIR);  
    while(pDirEnt != NULL)
    {
        if(    strcmp(pDirEnt->d_name, ".")  // not the dot
            && strcmp(pDirEnt->d_name, "..") // not the dot dot
          )
        {
            bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor, "%s                  "
                        , pDirEnt->d_name
                      );
            y_off += y_line;
            found_one = true;
        }
        pDirEnt = readdir(pDIR);
    }
    closedir(pDIR);    
    if(!found_one)
    {
        bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                    , "none found        "
                  );
        y_off += y_line;
    }    
    bmp_printf(   &BMP, x_off, y_off, BMP.white, bmp_bg_color, 0, font_size_factor
                , "                  "
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, BMP.white, bmp_bg_color, 0, font_size_factor
                , "                  "
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, BMP.white, bmp_bg_color, 0, font_size_factor
                , "                  "
              );
    y_off += y_line;
    bmp_to_SDL_Surface(&bmp, screen);
    return;
}

//############################################################################
void ez_uail_gui::display_directory_list(string directory)
{
    bool            found_one    = false;
    int             i,
                    y_line       = 8 * font_size_factor,
                    x_off        = BMP.yres + EZ_STATS_DISPLAY_OFFSET * font_size_factor,
                    y_off        = y_line;
    u_int           bmp_bg_color = BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b);
    DIR            *pDIR;
    struct dirent  *pDirEnt;

    pDIR = opendir(directory.c_str());
    if(pDIR == NULL)
        return;
    bmp_printf(   &BMP, x_off, 0, BMP.white, bmp_bg_color, 0, font_size_factor
                , "                  "
              );
    bmp_printf(   &BMP, x_off, y_off, BMP.white, bmp_bg_color, 0, font_size_factor
                , "                  "
              );
    bmp_printf(   &BMP, x_off, y_off, BMP.white, bmp_bg_color, 0, font_size_factor
                , "Palette set stats "
              );
    pDirEnt = readdir(pDIR);  
    while(pDirEnt != NULL)
    {
        if(!strstr(pDirEnt->d_name, ".")) // if there is no dot in the name
        {
            bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                        , "%s                  "
                        , pDirEnt->d_name
                      );
            y_off += y_line;
            found_one = true;
        }
        pDirEnt = readdir(pDIR);
    }
    closedir(pDIR);
    if(!found_one)
    {
        bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                    , "none found        "
                  );
        y_off += y_line;
    }
    bmp_printf(   &BMP, x_off, y_off, BMP.white, bmp_bg_color, 0, font_size_factor
                , "                  "
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, BMP.white, bmp_bg_color, 0, font_size_factor
                , "                  "
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, BMP.white, bmp_bg_color, 0, font_size_factor
                , "                  "
              );
    bmp_to_SDL_Surface(&bmp, screen);
    return;
}

//############################################################################
void ez_uail_gui::load_background_bitmap(struct ezfb_bitmap* bmp)
{
    bmp_flip(bmp);
    bmp_into_bmp(&BACKGROUND_BMP, bmp);
    show_background_bitmap = true;
    return;
}

//############################################################################
void ez_uail_gui::load_background_bitmap(string file)
{
    char file_name[81];
    struct ezfb_bitmap bmp;
    if(    file.size() < 4
        || file.substr(file.size() - 4) != ".bmp"
      )
        file += ".bmp";
    strcpy(file_name, (EZ_UAIL_BMP_SHARE + file).c_str());
    if(bmp_from_file(&bmp, file_name))
    {
        load_background_bitmap(&bmp);
        bmp_free(&bmp);
        bg_file_name = file;
    }
    return;
}

//############################################################################
void ez_uail_gui::clear_background_bitmap()
{
    bmp_uint(&BACKGROUND    , bg_color.as_uint());
    bmp_uint(&BACKGROUND_BMP, bg_color.as_uint());
    bg_file_name = ".";
    return;
}

//############################################################################
void ez_uail_gui::recolor_background()
{
    bmp_uint(&BACKGROUND    , bg_color.as_uint());
    bmp_uint(&BACKGROUND_BMP, bg_color.as_uint());
    if(bg_file_name != ".")
        load_background_bitmap(bg_file_name);
    return;
}

//############################################################################
void ez_uail_gui::display_frame_effects()
{
    int    i,
           y_line       = 8 * font_size_factor,
           x_off        = BMP.yres + EZ_STATS_DISPLAY_OFFSET * font_size_factor,
           y_off        = EZ_STATS_DISPLAY_OFFSET * font_size_factor;
    u_int  bmp_bg_color = BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b);
    
    bmp_printf(   &BMP, x_off, 0, BMP.white, bmp_bg_color, 0, font_size_factor
                , "                  "
              );
    bmp_printf(   &BMP, x_off, y_off, BMP.white, bmp_bg_color, 0, font_size_factor
                , "frame effects"
              );
    if(frame_effect_names.size() > 0)
    {
        for(i = 0; i < (int)frame_effect_names.size(); i++)
        {
            y_off += y_line;
            bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor, "%s          "
                        , frame_effect_names[i].c_str()
                      );
        }
    }
    else
    {
        y_off += y_line;
        bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                    , "no frame effects loaded"
                  );
    }        
    bmp_to_SDL_Surface(&BMP, screen);
    return;
}

//############################################################################
void ez_uail_gui::display_frame_set_effects()
{
    int    i,
           y_line       = 8 * font_size_factor,
           x_off        = BMP.yres + EZ_STATS_DISPLAY_OFFSET * font_size_factor,
           y_off        = EZ_STATS_DISPLAY_OFFSET * font_size_factor;
    u_int  bmp_bg_color = BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b);
    
    bmp_printf(   &BMP, x_off, 0, BMP.white, bmp_bg_color, 0, font_size_factor
                , "                  "
              );
    bmp_printf(   &BMP, x_off, y_off, BMP.white, bmp_bg_color, 0, font_size_factor
                , "frame set effects"
              );
    if(frame_set_effect_names.size() > 0)
    {
        for(i = 0; i < (int)frame_set_effect_names.size(); i++)
        {
            y_off += y_line;
            bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor, "%s          "
                        , frame_set_effect_names[i].c_str()
                      );
        }
    }
    else
    {
        y_off += y_line;
        bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                    , "no frame set effects loaded"
                  );
    }        
    bmp_to_SDL_Surface(&BMP, screen);
    return;
}

//############################################################################
//////////////////////////////////////////////////////////////////////////////
//############################################################################
